/*       ___   ___   ___   ___
 *      /_  \ /   \ /  _\ /  _\
 *     // \_/// \_/_\ / \\  / \\
 *    //  _ //  _ / \\\_// /  //
 *    \\_/ \\\_/ \\ //  /  \_//
 *     \___/ \___/ \\__/ \___/
 *     Transferts srie pour Thomson
 *
 *  Routines des commandes
 *
 *  Copyright (C) 2015 Puls
 *
 *  This program is frfp_logee software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>

#include "defs.h"

#ifdef DEBUG_MODE
    extern FILE *fp_log ;
    extern FILE *fp_map ;
#endif

extern struct CC90_T cc90 ;
extern struct DISK_T dk[4] ;
extern struct OPT_T opt ;
extern struct CONV_T conv ;

extern UBYTE *file_buffer ;
extern ULONG file_size ;
extern void FreeFile(void) ;

extern int GetFat(int) ;
extern int FindNext(int) ;
extern int FindFirst(int) ;
extern int FindClose(int) ;
extern int ReadSectors(int, UBYTE, UBYTE, UBYTE) ;
extern int SaveSectors(int, UBYTE, UBYTE, UBYTE) ;
extern int SendMemory(UBYTE, UBYTE *, USHORT, USHORT) ;
extern int DrawDirLine(int) ;
extern int DrawDiskLine(int) ;
extern int InitDrives(void) ;
extern int InitTrack20(int, int) ;
extern int LoadFile(int) ;
extern int SaveFile(int) ;
extern int DeleteThomsonEntry(int) ;
extern int SerialOpen(char *, int) ;
extern int SerialWrite(UBYTE *, size_t) ;
extern void SerialClose(void) ;

extern int ReadDirTrack(int) ;

extern void ParsePath(int, char *) ;
extern int Create(int, int) ;

extern int CreateThomsonName(char *, char *) ;

extern char /*@temp@*/*PathNameString(int, int) ;
extern int AskRequest(int, int) ;

extern void RawConsole(void) ;

extern int SendInfo(UBYTE, UBYTE, UBYTE, UBYTE, UBYTE, UBYTE) ;
extern int PError(char *, int) ;

/* Routines de conversion */
extern int TxtToAsm(void) ;
extern int TxtToAsc(void) ;
extern int AscToTxt(void) ;
extern int AsmToTxt(void) ;
extern int SndToWav(void) ;
extern int SonToWav(void) ;
extern int WavToSnd(void) ;
extern int WavToSon(void) ;
extern int MapToBmp(void) ;
extern int BmpToPpm(void) ;
extern int BmpToSnp(void) ;

extern void RawConsole(void) ;
extern void OriginalConsole(void) ;


/********************************************************
 * --install
 *     Installation sans disquette
 ********************************************************
 * Ct Thomson, le programme suivant doit avoir t lanc :
 *
 * 0 'SAVE"INSTALL"
 * 5 D=&H4000
 * 10 LOCATE,,0:CLS:CONSOLE,,1
 * 15 DATA "108EA7E0CE05001A"
 * 20 DATA "504F5FED22CC03FF"
 * 25 DATA "EDA4CC043CED22C6"
 * 30 DATA "01E7A48D1B1F988D"
 * 35 DATA "1733CB8D131F988D"
 * 40 DATA "0FE7C04A26F98D08"
 * 45 DATA "27E9E7A4ECC36ECB"
 * 50 DATA "E6A42BFC1F99E6A4"
 * 55 DATA "2BF63402C68031A4"
 * 60 DATA "A6A41E8848561024"
 * 65 DATA "FFF43582"
 * 70 A=D:FORI=1TO11:READA$:FORJ=1TO LEN(A$)-1 STEP2:POKEA,VAL
 * ("&H"+MID$(A$,J,2)):A=A+1:NEXTJ,I
 * 75 A=D/256:POKED+2,PEEK(D+2)+A:POKED+5,PEEK(D+5)+A:EXECD
 *
 * Pour le MO, remplacer la ligne 5 par :
 * 5 D=&H0000
 */
int SerialInstall(void)
{
    int success ;
    UBYTE  size = 0;
    USHORT offset = 0 ;
    USHORT mem_offset = 0 ;
    UBYTE *ptr = NULL ;
    UBYTE buf[6] ;

    /* Ouvre le port srie */
    if (SerialOpen(cc90.portname,cc90.timeout) == FALSE)
        return PError(cc90.portname,ERR_PORT_NOT_READY) ;
    dk[DRIVE1].cont = 4 ;
    dk[DRIVE1].nbtracks = 80 ;
    dk[DRIVE1].fatsize  = 160  ;
    dk[DRIVE1].secsize  = 256  ;

    /* Dclare le fichier 'CC90.BIN' de la disquette 'cc90.sap' */
    ParsePath(DRIVE1, "cc90.sap/CC90.BIN\0") ;

    /* Charge la piste de catalogue */
    if (ReadDirTrack(DRIVE1) == FALSE) return FALSE ;

    /* Charge le fichier CC90.BIN */
    switch(FindFirst(DRIVE1))
    {
        case FALSE :
            return FALSE ;
        case ERR_FILE_NOT_FOUND :
            (void)FindClose(DRIVE1) ;
            return PError(dk[DRIVE1].pattern,ERR_FILE_NOT_FOUND) ;
    }
    success = LoadFile(DRIVE1) ;
    (void)FindClose(DRIVE1) ;
    if (success == FALSE) return FALSE ;

    /* Envoie le fichier */
    if (file_buffer == NULL)
        return PError("",ERR_NOT_ENOUGH_MEMORY) ;
    ptr = file_buffer ;
    while(ptr[0] == 0x00)
    {
        /* Met  jour l'adresse relative d'implantation */
        offset = (USHORT)((*(ptr+3) << 8) | *(ptr+4)) ;
        buf[0] = (UBYTE)((offset - mem_offset) >> 8) ;
        buf[1] = (UBYTE)(offset - mem_offset) ;
        /* Met  jour la taille du bloc */
        size = *(ptr+2) ;
        buf[2] = size ;
        if (SerialWrite(buf,3) == FALSE) return FALSE ;
        /* Envoie les donnes du bloc */
        if (SerialWrite(&ptr[5],size) == FALSE) return FALSE ;
        /* Envoie le flag pour bloc suivant */
        buf[0] = 0x00 ;
        if (SerialWrite(buf,1) == FALSE) return FALSE ;

#ifdef DEBUG_MODE
        int i ;
        fprintf(fp_log,"SerialInstall = [size = %d][offset = $%04x][index = $%04x] ",
                            size,
                            offset,
                            offset - mem_offset) ;
        fprintf(fp_log,"\n") ;
        for(i=0;i<size;i++)
        {
            if ((i & 0x0F) == 0)
            {
                if (i) fprintf(fp_log,"\n") ;
                fprintf(fp_log,"%04x  ",offset+i) ;
            }
            fprintf(fp_log,"%02x ",ptr[i+5] & 0xFF) ;
        }
        fprintf(fp_log,"\n") ;
#endif
#ifdef _GNU_SOURCE
        usleep(40000) ;
#endif
        ptr += 5 + size ;
        mem_offset = offset + size ;
    }

    /*
     * Envoie la demande d'excution du programme
     */
    /* Rcupre l'adresse d'excution du binaire */
    offset = (USHORT)((*(ptr+3) << 8) | *(ptr+4)) ;
    /* Adresse relative d'implantation (0x0000) */
    buf[0] = 0x00 ;
    buf[1] = 0x00 ;
    /* Taille des donnes */
    buf[2] = 0x02 ;
    /* Adresse relative d'excution */
    buf[3] = (UBYTE)((offset - mem_offset) >> 8) ;
    buf[4] = (UBYTE)(offset - mem_offset) ;
    /* Fin de transfert (CTS  1) */
    buf[5] = 0x03 ;
    if (SerialWrite(buf,6) == FALSE) return FALSE ;

#ifdef DEBUG_MODE
    fprintf(fp_log,"SerialInstall (exec) = [offset = $%04x][index = $%04x] ",
                   offset,
                   mem_offset) ;
#endif
    SerialClose() ;
    FreeFile() ;

    return TRUE ;
}


/********************************************************
 * --create
 *     Cration d'un disque virtuel
 ********************************************************/
int CreateDisk(void)
{
    return Create(DRIVE0, dk[DRIVE0].cont) ;
}


/********************************************************
 * --format
 *     Formatage d'un disque Thomson
 ********************************************************/
int FormatDisk(void)
{
    UBYTE track ;

    /* Si formatage lent */
    if(opt.quick == FALSE)
    {
        switch(dk[DRIVE0].type)
        {
            case TO_TYPE :
                if (SendInfo(CODE_FORMATDISK, dk[DRIVE0].drive, (UBYTE)opt.interlace, 0, 0, 0) == FALSE)
                    return FALSE ;
                break ;
            case DCMOTO_TYPE :
            case SAP_TYPE :
                for(track=0;track<dk[DRIVE0].nbtracks;track++)
                    if (SaveSectors(DRIVE0, track, FIRSTB1, LASTB2) == FALSE)
                        return FALSE ;
                if (InitTrack20(DRIVE0, dk[DRIVE0].cont) == FALSE)
                        return FALSE ;
                break ;
            case FILE_TYPE :
                return PError("",ERR_ILLEGAL_OPERATION) ;
            default :
                return PError(dk[DRIVE0].arg,ERR_CANT_OPEN_FILE) ;
        }
    }
    else
    /* Si formatage rapide */
       if (InitTrack20(DRIVE0, dk[DRIVE0].cont) == FALSE)
           return FALSE ;

    return TRUE ;
}


/********************************************************
 * --delete
 *     Effacement de fichiers Thomson
 ********************************************************/
static void DrawDeleteCounter(int filecounter)
{
    char plural_sign[2] = "\0" ;
    if (filecounter > 1) strcat(plural_sign,"s") ;
    (void)printf("  === %d fichier%s efface%s\n",filecounter,plural_sign,plural_sign) ;
}

int DeleteFiles(void)
{
    int pattern = FALSE ;
    int success = TRUE ;
    int filefound ;
    int filecounter = 0 ;

    /* N'autorise pas l'effacement des fichiers PC */
    switch(dk[DRIVE0].type)
    {
        case FILE_TYPE :
        case INVALID_TYPE :
            return PError("",ERR_ILLEGAL_OPERATION) ;
    }

    /* Repre un pattern */
    if (strpbrk(dk[DRIVE0].pattern,"*?"))
        pattern = TRUE ;

    /* Charge la piste de catalogue */
    if (ReadDirTrack(DRIVE0) == FALSE) return FALSE ;

    /* Recherche premire correspondance */
    switch(filefound = FindFirst(DRIVE0))
    {
        case FALSE :
        case ERR_FILE_NOT_FOUND :
            (void)FindClose(DRIVE0) ;
            DrawDeleteCounter(filecounter) ;
            return FALSE ;
    }
    do
    {
        if (filefound != ERR_FILE_IS_A_DIR)
        {
            success = TRUE ;
            if (AskRequest(DRIVE0,DELETE_FLAG) == TRUE)
            {
                (void)printf("Effacement de %s\n",PathNameString(FALSE, DRIVE0)) ;
                /* Transfert du fichier */
                switch(dk[DRIVE0].type)
                {
                    case TO_TYPE :
                    case SAP_TYPE :
                    case DCMOTO_TYPE :
                        if ((success = DeleteThomsonEntry(DRIVE0)) == TRUE)
                            filecounter++ ;
                        break ;
                    default :
                        success = PError("",ERR_ILLEGAL_OPERATION) ;
                        break ;
                }
            }
        }
        filefound = FindNext(DRIVE0) ;
    } while ((pattern == TRUE)
          && ((success == TRUE) || (cc90.lasterror > ____DONT_STOP))
          && (cc90.exit == FALSE)
          && ((filefound == TRUE) || (filefound == ERR_FILE_IS_A_DIR))) ;
    (void)FindClose(DRIVE0) ;

    DrawDeleteCounter(filecounter) ;
    return TRUE ;
}


/********************************************************
 * --mem
 *     Implantation mmoire d'un fichier binaire Thomson
 ********************************************************/
int SendBinary(void)
{
    int i = 0 ;
    USHORT size ;
    static char tofilename[13] ;

    /* Le drive 1 est dclar en drive srie
     * pour en mme temps ouvrir le port srie
     * et envoyer le bloc Ask */
    dk[DRIVE1].type = TO_TYPE ;
    (void)printf("Implantation de '%s'\n",dk[DRIVE0].pattern) ;

    /* Initialisation des lecteurs */
    if (InitDrives() == FALSE) return FALSE ;

    /* Charge la piste de catalogue */
    if (ReadDirTrack(DRIVE0) == FALSE) return FALSE ;

    /* Chargement du fichier BIN */
    switch(FindFirst(DRIVE0))
    {
        case FALSE :
            return FALSE ;
        case ERR_FILE_NOT_FOUND :
            (void)FindClose(DRIVE0) ;
            return PError(dk[DRIVE0].pattern,ERR_FILE_NOT_FOUND) ;
    }
    if (LoadFile(DRIVE0) == FALSE) return FALSE ;
    (void)FindClose(DRIVE0) ;

    /* Vrifie la validit du fichier */
    dk[DRIVE0].filename[0] = '\0' ;
    strncat(dk[DRIVE0].filename,dk[DRIVE0].pattern,STRING_SIZE_MAX-1) ;
    (void)CreateThomsonName(dk[DRIVE0].filename,tofilename) ;
    if (strncmp(&tofilename[8],"BIN",3) != 0)
        return PError(dk[DRIVE0].filename,ERR_BAD_FILE_FORMAT) ;

    /* Envoie les donnes du fichier binaire au Thomson */
    if (file_buffer == NULL)
        return PError("",ERR_NOT_ENOUGH_MEMORY) ;
    while(file_buffer[i] == 0x00)
    {
        size = ((USHORT)*(file_buffer+i+1) << 8) | (USHORT)*(file_buffer+i+2) ;
        if (SendMemory(opt.bank,
                       file_buffer+i+5,
                       ((USHORT)*(file_buffer+i+3) << 8) | (USHORT)*(file_buffer+i+4),
                       size) == FALSE)
            return FALSE ;

        i += 5 + (int)size ;
        if ((size_t)i >= file_size)
            return PError(dk[DRIVE0].filename,ERR_BAD_FILE_FORMAT) ;
    }

    /* Demande ventuellement l'excution du programme */
    if ((((USHORT)*(file_buffer+i+3) << 8) | (USHORT)*(file_buffer+i+4)) != (USHORT)0x0000)
        if (SendInfo(CODE_SENDMEMORY, opt.bank,
                                      *(file_buffer+i+3), *(file_buffer+i+4),
                                      0, 0) == FALSE)
            return FALSE ;

    FreeFile() ;

    return TRUE ;
}


/********************************************************
 * --dir
 *     Affichage d'un rpertoire
 ********************************************************/
static void DrawDirCounter(int filecounter)
{
    char plural_sign[2] = "\0" ;
    if (filecounter > 1) strcat(plural_sign,"s") ;
    (void)printf("  === %d fichier%s\n",filecounter,plural_sign) ;
}

int DrawDir(void)
{
    int success ;
    int filefound ;
    int filecounter = 0 ;

    /* Charge la piste de catalogue */
    if (ReadDirTrack(DRIVE0) == FALSE) return FALSE ;

    /* Recherche premire correspondance */
    filefound = FindFirst(DRIVE0) ;  
    /* Affiche en-tte du directory
     * L'affichage de l'en-tte du directory
     * se fait  cet instant car si le rpertoire
     * ne comporte aucun fichier, au moins les
     * informations sur la disquette seront
     * affiches */
    if (DrawDiskLine(DRIVE0) == FALSE)
        return FALSE ;
    /* Traite premire correspondance */
    switch(filefound)
    {
        case FALSE :
        case ERR_FILE_NOT_FOUND :
            (void)FindClose(DRIVE0) ;
            DrawDirCounter(filecounter) ;
            return FALSE ;
    }

    do
    {
        success = TRUE ;
        if (filefound == ERR_FILE_IS_A_DIR)
            (void)printf("  (dir) %s\n",dk[DRIVE0].filename) ;
        else
        {
            if ((success = DrawDirLine(DRIVE0)) == TRUE)
                filecounter++ ;
        }
        filefound = FindNext(DRIVE0) ;
    } while ((success == TRUE)
          && ((filefound == TRUE) || (filefound == ERR_FILE_IS_A_DIR))) ;
    (void)FindClose(DRIVE0) ;
    DrawDirCounter(filecounter) ;
    return TRUE ;
}


/********************************************************
 * --file
 *     Copie d'un fichier
 ********************************************************/
static void DrawCopyCounter(int filecounter)
{
    char plural_sign[2] = "\0" ;
    if (filecounter > 1) strcat(plural_sign,"s") ;
    (void)printf("  === %d fichier%s copie%s\n",filecounter,plural_sign,plural_sign) ;
}


int
CopyFile(void)

{
    int i ;
    int filecounter = 0 ;
    int pattern = FALSE ;
    int success = TRUE ;
    int filefound ;
    int (*prog)(void) ;
    char *found ;
    char suffix[6] ;
    int empty = (dk[DRIVE1].pattern[0] == '\0') ? TRUE : FALSE ;

    const struct {
        int  *reg ;
        char pc_suffix[5] ;
        struct {
            int  value ;
            char to_suffix[5] ;
            int  (*to_prog)(void) ;
            int  (*pc_prog)(void) ;
        } list[2] ;
    } conv_table[3] = {
        { &conv.txttype, ".txt", 
            { { CONV_TXT_ASM, ".ASM", TxtToAsm, AsmToTxt }, 
              { CONV_TXT_ASC, ".ASC", TxtToAsc, AscToTxt } } },
        { &conv.sndtype, ".wav",
            { { CONV_SND_SON, ".SON", WavToSon, SonToWav },
              { CONV_SND_SND, ".SND", WavToSnd, SndToWav } } },
        { &conv.maptype, ".bmp",
            { { CONV_GFX_SNP, ".MAP", BmpToSnp, MapToBmp },
              { CONV_GFX_PPM, ".PPM", BmpToPpm, MapToBmp } } }
    } ;

    /* Lit la piste de catalogue */
    if (ReadDirTrack(DRIVE0) == FALSE) return FALSE ;
    if (ReadDirTrack(DRIVE1) == FALSE) return FALSE ;

    /* Repre un pattern */
    if (strpbrk(dk[DRIVE0].pattern,"*?"))
        pattern = TRUE ;
    else
    /* Actualise ventuellement le nom de fichier pour sauvegarde */
    if (dk[DRIVE1].pattern[0] == '\0')
        strcat(dk[DRIVE1].pattern,dk[DRIVE0].pattern) ;

    /* Recherche premire correspondance */
    switch(filefound = FindFirst(DRIVE0))
    {
        case FALSE :
        case ERR_FILE_NOT_FOUND :
            (void)FindClose(DRIVE0) ;
            DrawCopyCounter(filecounter) ;
            return FALSE ;
    }
    do
    {
        if (filefound != ERR_FILE_IS_A_DIR)
        {
            success = TRUE ;
            if (dk[DRIVE1].pattern[0] == '\0')
                strcpy(dk[DRIVE1].pattern,dk[DRIVE0].filename) ;
            if ((dk[DRIVE0].dir == TRUE) && (dk[DRIVE1].dir == FALSE))
                success = PError(dk[DRIVE1].arg,ERR_COPY_DIR_INTO_FILE) ;
            else
            {
                /* Reprage de la conversion */
                prog = NULL ;
                if ((conv.convert == TRUE)
                 && ((found = strrchr(dk[DRIVE0].filename,'.')) != NULL))
                {
                    /* Rcupre le suffixe du fichier source */
                    strncpy(suffix,found,5) ;

                    /* Repre le suffixe du fichier destination */
                    if ((found = strrchr(dk[DRIVE1].pattern,'.')) == NULL)
                        found = dk[DRIVE1].pattern + strlen(dk[DRIVE1].pattern) ;

                    /* Recherche les caractristiques de conversion dans la
                     * table de conversion, remplace ventuellement le suffixe
                     * destination et repre le vecteur de routine */
                    i = 0 ;
                    while ((i < 3) && (prog == NULL))
                    {
                        /* Si suffixe de fichier PC */
                        if (strcmp(suffix,conv_table[i].pc_suffix) == 0)
                        {
                            if (strcmp(found,conv_table[i].list[0].to_suffix) == 0)
                                prog = conv_table[i].list[0].to_prog ;
                            else
                            if (strcmp(found,conv_table[i].list[1].to_suffix) == 0)
                                prog = conv_table[i].list[1].to_prog ;
                            else
                            {
                                if ((found[0] == '\0') || (empty == TRUE))
                                    strcpy(found,conv_table[i].list[*conv_table[i].reg].to_suffix) ;
                                prog = conv_table[i].list[*conv_table[i].reg].to_prog ;
                            }
                        }
                        else
                        /* Si suffixe de fichier Thomson */
                        if (strcmp(suffix,conv_table[i].list[0].to_suffix) == 0)
                        {
                            if ((found[0] == '\0') || (empty == TRUE))
                                strcpy(found,conv_table[i].pc_suffix) ;
                            prog = conv_table[i].list[0].pc_prog ;
                        }
                        else
                        if (strcmp(suffix,conv_table[i].list[1].to_suffix) == 0)
                        {
                            if ((found[0] == '\0') || (empty == TRUE))
                                strcpy(found,conv_table[i].pc_suffix) ;
                            prog = conv_table[i].list[1].pc_prog ;
                        }
                        i++ ;
                    }
                }
                strcpy(dk[DRIVE1].filename,dk[DRIVE1].pattern) ;

                /* Affiche message de copie */
                (void)printf("Copie de %s",PathNameString(TRUE, DRIVE0)) ;
                (void)printf(" vers %s\n",PathNameString(FALSE, DRIVE1)) ;

                /* Requte d'crasement */
                success = FindFirst(DRIVE1) ;
                if ((success == ERR_FILE_NOT_FOUND)
                 || ((success == TRUE) && (AskRequest(DRIVE1,OVERWRITE_FLAG) == TRUE)))
                {
#ifdef DEBUG_MODE
                    fprintf(fp_map,"\n\n\n%s\n\n\n",dk[1].filename) ;
#endif
                    /* Chargement, conversion ventuelle et sauvegarde du fichier */
                    if (((success = LoadFile(DRIVE0)) == TRUE) && (prog != NULL))
                        success = (*prog)() ;
                    if ((success == TRUE) && ((success = SaveFile(DRIVE1)) == TRUE))
                        filecounter++ ;
                }
            }
            FreeFile() ;
        }
        dk[DRIVE1].pattern[0] = '\0' ;
        filefound = FindNext(DRIVE0) ;
    } while ((pattern == TRUE)
          && ((success == TRUE) || (cc90.lasterror > ____DONT_STOP))
          && (cc90.exit == FALSE)
          && ((filefound == TRUE) || (filefound == ERR_FILE_IS_A_DIR))) ;
    (void)FindClose(DRIVE0) ;

    DrawCopyCounter(filecounter) ;
    return TRUE ;
}


/********************************************************
 * --disk
 *     Copie d'un disque
 ********************************************************/
int
CopyDisk(void) /*@globals errno,stdout@*/
{
    int drive ;
    UBYTE track, first, last ;

    /* Vrification de la concordance et de la validit de tous les formats */
    for(drive = 0; drive < 4; drive++)
        if (dk[drive].type != INVALID_TYPE)
            if ((dk[drive].type == FILE_TYPE)
             || (dk[drive].nbtracks != dk[DRIVE0].nbtracks)
             || (dk[drive].secsize != dk[DRIVE0].secsize))
                return PError(dk[drive].dkname,ERR_INCOMPATIBLE_DISK) ;

    /* Vrifie qu'il n'y a pas de mlange de contrleur interne/externe */
    for(drive = 0; drive < 2; drive++)
        if ((dk[drive].type != INVALID_TYPE)
         && (dk[drive+2].type != INVALID_TYPE)
         && (dk[drive].external != dk[drive+2].external))
            return PError(dk[drive+2].dkname,ERR_INCOMPATIBLE_DISK) ;

    /* Chargement ventuel de la FAT  pour copie rapide */
    if (opt.quick == TRUE)
        for(drive=0; drive<4; drive+=2)
            if (dk[drive].type != INVALID_TYPE)
                if (GetFat(drive) == FALSE)
                    return FALSE ;

    /* Transfert des blocs */
    for(track = 0; track < dk[DRIVE0].nbtracks; track++)
    {
        for(drive = 0 ; drive < 4 ; drive += 2)
        {
            if (dk[drive].type != INVALID_TYPE)
            {
                (void)printf("\rPiste %d ",(int)track) ;
                (void)fflush(stdout) ;

                /* Limite les secteurs de dpart et d'arrive */
                first = ((dk[drive].fat[track*2] == 0xFF) && (opt.quick == TRUE)) ? FIRSTB2 : FIRSTB1 ;
                last = ((dk[drive].fat[track*2+1] == 0xFF) && (opt.quick == TRUE)) ? LASTB1 : LASTB2 ;

                /* Transfert conditionnel du groupe de secteurs */
                if (first < last)
                {
                    if (ReadSectors(drive, track, first, last) == FALSE) return FALSE ;
                    if (SaveSectors(drive+1, track, first, last) == FALSE) return FALSE ;
                }
            }
        }
    }
    (void)printf("\n") ;
    return TRUE ;
}
