/*       ___   ___   ___   ___
 *      /_  \ /   \ /  _\ /  _\
 *     // \_/// \_/_\ / \\  / \\
 *    //  _ //  _ / \\\_// /  //
 *    \\_/ \\\_/ \\ //  /  \_//
 *     \___/ \___/ \\__/ \___/
 *     Transferts srie pour Thomson
 *
 *  Conversion de fichiers ASC<->TXT
 *
 *  Copyright (C) 2015 Puls
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <string.h>

#include "defs.h"

extern int PError(char *, int) ;

extern struct CONV_T conv ;
extern struct DISK_T dk[4] ;

extern UBYTE *file_buffer ;
extern ULONG file_size ;
extern UBYTE *temp_buffer ;
extern ULONG temp_size ;
extern void AllocTemp(size_t) ;
extern void IncreaseTempBuffer(size_t) ;
extern void FreeTemp(size_t) ;

/*
 * Liste des correspondances ISO 8859-1
 * Le premier caractre est l'incrment de compteur de caractres.
 * ATTENTION !!! Il ne s'agit pas du nombre de caractres de la
 * srie, mais de l'occupation de cette srie de caractres sur l'cran.
 * Les trois derniers donnent la correspondance Thomson approprie.
 * Les 0 (s'il y en a) indiquent la fin de la liste de codes
 * pour chaque caractre. Sinon, le nombre de caractres 
 * crire se limite  3. */
struct TXT_TO_ASC_TABLE {
    int    size ;
    char   list[4] ;
} ;
static const struct TXT_TO_ASC_TABLE txt_to_asc_t[96] = {
    { 1, " "       }, { 1, "!"       }, { 1, "c"       }, { 1, "\22#"    },
    { 1, "*"       }, { 1, "Y"       }, { 1, ":"       }, { 1, "\22'"    },
    { 1, "\x27"    }, { 3, "(c)"     }, { 1, "\22\x30" }, { 2, "<<"      },
    { 1, "-"       }, { 1, "-"       }, { 3, "(r)"     }, { 1, "\126"    },
    { 1, "\22\x30" }, { 1, "\22\x31" }, { 2, "^2"      }, { 2, "^3"      },
    { 1, "'"       }, { 1, "u"       }, { 1, "q"       }, { 1, "."       },
    { 1, ","       }, { 2, "^1"      }, { 1, "o"       }, { 2, ">>"      },
    { 1, "\22<"    }, { 1, "\22="    }, { 1, "\22>"    }, { 1, "?"       },
    { 1, "\22AA"   }, { 1, "\22BA"   }, { 1, "\22CA"   }, { 1, "A"       },
    { 1, "\22HA"   }, { 1, "A"       }, { 2,  "AE"     }, { 1, "C"       },
    { 1, "\22AE"   }, { 1, "\22BE"   }, { 1, "\22CE"   }, { 1, "\22HE"   },
    { 1, "\22AI"   }, { 1, "\22BI"   }, { 1, "\22CI"   }, { 1, "\22HI"   },
    { 1, "D"       }, { 1, "N"       }, { 1, "\22AO"   }, { 1, "\22BO"   },
    { 1, "\22CO"   }, { 1, "O"       }, { 1, "\22HO"   }, { 1, "x"       },
    { 1, "0"       }, { 1, "\22AU"   }, { 1, "\22BU"   }, { 1, "\22CU"   },
    { 1, "\22HU"   }, { 1, "\22BY"   }, { 1, "p"       }, { 1, "\22\x7b" },
    { 1, "\22Aa"   }, { 1, "\22Ba"   }, { 1, "\22Ca"   }, { 1, "a"       },
    { 1, "\22Ha"   }, { 1, "a"       }, { 2, "ae"      }, { 1, "\22Kc"   },
    { 1, "\22Ae"   }, { 1, "\22Be"   }, { 1, "\22Ce"   }, { 1, "\22He"   },
    { 1, "\22Ai"   }, { 1, "\22Bi"   }, { 1, "\22Ci"   }, { 1, "\22Hi"   },
    { 1, "o"       }, { 1, "n"       }, { 1, "\22Ao"   }, { 1, "\22Bo"   },
    { 1, "\22Co"   }, { 1, "o"       }, { 1, "\22Ho"   }, { 1, "\22\x38" },
    { 1, "0"       }, { 1, "\22Au"   }, { 1, "\22Bu"   }, { 1, "\22Cu"   },
    { 1, "\22Hu"   }, { 1, "\22By"   }, { 1, "P"       }, { 1, "\22Hy"   }
} ;

/*
 * Traitement d'un fichier ASC (PC->TO)
 */
int TxtToAsc(void)
{
    int c ;
    int i ;
    int spacetab = 0 ;   /* Position dans la tabulation */
    int teletel = 0 ;    /* Flag Tltel */
    int keymap_index = 0 ;
    UBYTE *src ;
    UBYTE *dst ;

    AllocTemp(1024) ;
    if ((temp_buffer == NULL) || (file_buffer == NULL))
        return PError("", ERR_NOT_ENOUGH_MEMORY) ;
    src = file_buffer ;
    dst = temp_buffer ;

    while ((size_t)(src-file_buffer) < file_size)
    {
        /* Lecture du caractre */
        c = (int)*(src++) ;

        switch(c)
        {
            case 0x09 :
                if (conv.tabul == TRUE)
                {
                    spacetab = conv.tabul - (spacetab % conv.tabul) ;
                    for (i=0;i<spacetab;i++)
                        *(dst++) = 0x20 ;
                }
                else
                    *(dst++) = 0x09 ;
                break ;

            case 0x0D :
            case 0x00 :
                /* Repasse en mode normal et CR+LF */
                if (teletel == 1) *(dst++) = 0x0F ;
                teletel = 0 ;
                spacetab = 0 ;
                *(dst++) = 0x0D ;
                *(dst++) = 0x0A ;
                break ;

            default :
                /* Si caractre > 127 */
                if (c > 0x7F)
                {
                    if (c >= 0xA0)
                    {
                        keymap_index = c - 0xA0 ;
                        spacetab += txt_to_asc_t[keymap_index].size ;
                        /* Dsactive le mode Tltel */
                        if (teletel == 1)
                        {
                            teletel = 0 ;
                            *(dst++) = 0x0F ;
                        }
                        /* Ecrit la srie de caractres */
                        i = 0 ;
                        while (txt_to_asc_t[keymap_index].list[i] != '\0')
                            *(dst++) = (UBYTE)txt_to_asc_t[keymap_index].list[i++];
                    }
                }
                else
                /* Si caractre 32-127 */
                if (c >= 0x20)
                {
                    /* Ecrit le caractre */
                    spacetab++ ;
                    *(dst++) = (UBYTE)c ;
                }
                break ;
        }

        /* Agrandi le buffer si ncessaire */
        if ((size_t)(dst-temp_buffer) > (temp_size - 20))
        {
            dst -= (size_t)temp_buffer ;
            IncreaseTempBuffer(1024) ;
            dst += (size_t)temp_buffer ;
        }
    }

    /* Repasse en mode normal si mode tltel */
    if (teletel == 1) *(dst++) = 0x0F ;

    /* Officialisation des nouvelles adresse et taille de fichier */
    FreeTemp((size_t)(dst-temp_buffer)) ;

    return TRUE ;
}

/* Table pour ESC 3 (squence 3)
 *
 * Pour chaque jeu :
 * grave, aigu, circ, trema, cedille
 */
static const char table_esc_3[] = {
    '\xc0', '\xc1', '\xc2', '\xc4', 'A' ,
    '\xc8', '\xc9', '\xca', '\xcb', 'E' ,
    '\xcc', '\xcd', '\xce', '\xcf', 'I' ,
    '\xd2', '\xd3', '\xd4', '\xd6', 'O' ,
    '\xd9', '\xda', '\xdb', '\xdc', 'U' ,
    '\xdd', 'Y' , 'Y' , 'Y' , 'Y' ,
    'C' , 'C' , 'C' , 'C' , '\xc7',
    '\xe0', '\xe1', '\xe2', '\xe4', 'a' ,
    '\xe8', '\xe9', '\xea', '\xeb', 'e' ,
    '\xec', '\xed', '\xee', '\xef', 'i' ,
    '\xf2', '\xf3', '\xf4', '\xf6', 'o' ,
    '\xf9', '\xfa', '\xfb', '\xfc', 'u' ,
    '\xfd', 'y' , 'y' , 'y' , 'y' ,
    'c' , 'c' , 'c' , 'c' , '\xe7'
} ;

/* Table pour ESC 2 (squence 2) */

static char tab_asc1[] = "ABCHK#$&,-./018<=>jz{'" ;

static char tab_asc4[][3] = {
    "\xa3","$","#","<-","^","->","v","\xb0","\xb1",
    "\xf7","\xbc","\xbd","\xbe","\x8c","\x9c","\xdf","\xa7"
} ;


/* Table pour correspondance Tltel */
static char teletel_t[5] = {
      '\x20',  /* pave */
      '\x20',  /* pave bas */
      '\x20',  /* pave gauche */
      '\x20',  /* pave droite */
      '\x20'   /* pave haut */
} ;

static char special_ascii[3] = { '\x2d', '\xaf', '\x81' } ;


/*
 * Traitement d'un fichier ASC (TO->PC)
 */
int AscToTxt(void)
{
    /* Table pour ESC 3 (squence 1) */

    /* Table pour ESC 3 (squence 2) */
    char tab_asc2[] = "AEIOUYCaeiouyc" ;

    /* Table pour caractres Tltel */
    char teletel_ref[]= {
        '\x7f',    /* pave */
        '\x70',    /* pave bas */
        '\x35',    /* pave gauche */
        '\x6a',    /* pave droite */
        '\x23'     /* pave haut */
    };

    int c ;
    int i, j, count = 0 ;
    int spacetabs = 0;
    int teletel = FALSE ;
    UBYTE *src ;
    UBYTE *dst ;

    /* Alloue la mmoire pour la conversion du fichier */
    AllocTemp(1024) ;
    if ((temp_buffer == NULL) || (file_buffer == NULL))
        return PError("", ERR_NOT_ENOUGH_MEMORY) ;
    src = file_buffer ;
    dst = temp_buffer ;

    while ((size_t)(src-file_buffer) < file_size)
    {
        /* Lecture du caractre */
        c = (int)*(src++) ;

        switch(c)
        {
            case 0x09 :
                if (conv.tabul == TRUE)
                    for(i=0;i<(conv.tabul-(spacetabs % conv.tabul));i++)
                        *(dst++) = 0x20 ;
                else
                     *(dst++) = 0x09 ;
                spacetabs = 0 ;
                break ;

            case 0x0D :
            case 0x00 :
                count = 0 ;
                spacetabs = 0 ;
                *(dst++) = 0x0D ;
                *(dst++) = 0x0A ;
                break ;

            case 0x0E :
                teletel = TRUE ;
                break ;

            case 0x0F :
                teletel = FALSE ;
                break ;

            case 0x16 :
                c = (int)*(src++) ;  /* Lit le caractre */
                i = 0 ;
                while ((i<(int)strlen(tab_asc1)) && (c != (int)tab_asc1[i]))
                    i++ ;
                if (i<5)  /* Si accentuation */
                {
                    c = (int)*(src++) ;
                    j = 0 ;
                    while ((j<(int)strlen(tab_asc2)) && (c != (int)tab_asc2[j]))
                        j++ ;
                    if (j<(int)strlen(tab_asc2)) {
                        *(dst++) = (UBYTE)table_esc_3[(((conv.keymap_number * 14)+j) *5) + i] ;
                        spacetabs++ ; }
                }
                else
                    if (i<(int)strlen(tab_asc1))  /* Si caractre G3 existe */
                        for(j=0;j<2;j++)  /* Ecrit srie de caractres */
                            if (tab_asc4[(conv.keymap_number * 17)+i-5][j] != '\0') {
                                *(dst++) = (UBYTE)tab_asc4[(conv.keymap_number * 17)+i-5][j] ;
                                spacetabs++ ;}
                break ;

            default :
                if (c > 0x7F) c = 0x20 ;
                if (c >= 0x20)
                {
                    if (teletel == TRUE)
                    {
                        /* Recherche dans la liste */
                        i = 0 ;
                        while ((i<(int)sizeof(teletel_ref)) && (c != (int)teletel_ref[i]))
                            i++ ;
                        /* Si lettre existe */
                        c = (i<(int)sizeof(teletel_ref))
                                ? (int)teletel_t[i]
                                : 0x20 ;
                    }
                    else
                    /* Si caractre affichable */
                    {
                        /* Conversions spciales */
                        if (c == 0x60)  c = (int)special_ascii[0] ;
                        if (c == 0x7E)  c = (int)special_ascii[1] ;
                        if (c == 0x7F)  c = (int)special_ascii[2] ;
                    }
                    spacetabs++ ;
                    *(dst++) = (UBYTE)c ;
                }
                break ;
        }

        count++ ;
        if (count >= 80)
        {
            count = 0 ;
            spacetabs = 0 ;
            *(dst++) = 0x0D ;
            *(dst++) = 0x0A ;
        }

        /* Agrandi le buffer si ncessaire */
        if ((size_t)(dst-temp_buffer) > (temp_size - 20))
        {
            dst -= (size_t)temp_buffer ;
            IncreaseTempBuffer(1024) ;
            dst += (size_t)temp_buffer ;
        }
    }

    FreeTemp((size_t)(dst-temp_buffer)) ;

    return TRUE ;
}
