/*       ___   ___   ___   ___
 *      /_  \ /   \ /  _\ /  _\
 *     // \_/// \_/_\ / \\  / \\
 *    //  _ //  _ / \\\_// /  //
 *    \\_/ \\\_/ \\ //  /  \_//
 *     \___/ \___/ \\__/ \___/
 *     Transferts srie pour Thomson
 *
 *  Conversion de fichiers ASM<->TXT
 *
 *  Copyright (C) 2015 Puls
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <string.h>

#include "defs.h"

extern struct CONV_T conv ;

extern int PError(char *, int) ;

extern UBYTE *file_buffer ;
extern ULONG file_size ;
extern UBYTE *temp_buffer ;
extern ULONG temp_size ;
extern void AllocTemp(size_t) ;
extern void IncreaseTempBuffer(size_t) ;
extern void FreeTemp(size_t) ;

static char txt_to_asm_t[128] = {
         'e', CHAR127,     ',',     'f',     ',',     '.',     '+',     '+',    /* 80-87 */
         '^',     '%',     'S',     '<',     'O', CHAR127,     'Z', CHAR127,    /* 88-8F */
     CHAR127,    '\'',    '\'',     '"',     '"',     '.',     '-',     '-',    /* 90-97 */
         '-',     'T',     's',     '>',     'o', CHAR127,     'z',     'Y',    /* 98-9F */
         ' ',     '!',     'c',     'L',     '*',     'Y',     ':',     'p',    /* A0-A7 */
         '"',     'c',     'a',     '<',     '-',     '-',     'r',     '-',    /* A8-AF */
         '0',     '+',     '2',     '3',    '\'',     'u',     'q',     '.',    /* B0-B7 */
         ',',     '1',     '0',     '>',     '/',     '/',     '/',     '?',    /* B8-BF */
         'A',     'A',     'A',     'A',     'A',     'A',     'A',     'C',    /* C0-C7 */
         'E',     'E',     'E',     'E',     'I',     'I',     'I',     'I',    /* C8-CF */
         'D',     'N',     'O',     'O',     'O',     'O',     'O',     'x',    /* D0-D7 */
         '0',     'U',     'U',     'U',     'U',     'Y',     'p',     'B',    /* D8-DF */
      '\x84',  '\x81',  '\x82',     'a',  '\x83',     'a',     'a',  '\x80',    /* E0-E7 */
      '\x88',  '\x85',  '\x86',  '\x87',     'i',     'i',  '\x89',  '\x8a',    /* E8-EF */
         'o',     'n',     'o',     'o',  '\x8b',     'o',  '\x8c',     ':',    /* F0-F7 */
         '0',  '\x8f',     'u',  '\x8d',  '\x8e',     'y',     'P',     'y'     /* F8-FF */
} ;


/*
 * Traitement d'un fichier ASM (PC->TO)
 */

int TxtToAsm(void)
{
    UBYTE c ;
    int i0 = 0, i1 = 0 ;
    UBYTE *src ;
    UBYTE *dst ;

    AllocTemp(1024) ;
    if ((temp_buffer == NULL) || (file_buffer == NULL))
        return PError("", ERR_NOT_ENOUGH_MEMORY) ;
    src = file_buffer ;
    dst = temp_buffer ;

    while ((size_t)(src-file_buffer) < file_size)
    {
        c = *(src++) ;
        switch(c)
        {
            /* Tabulations ASM */
            case 0x09 :
                i1 += (i1 < 22) ? 7 + ((i1 >=14 ) ? 1 : 0) - (i1 % 7) : 1 ; 
                break ;

            /* CR ou NULL */
            case 0x0A :
            case 0x00  :
                *(dst++) = 0x0D ;
                i1 = 0 ;
                i0 = 0 ;
                break ;

           /* Espace */
           case 0x20 :
                i1++ ;
                break ;

           default :
               /* Si > 32 */
               if (c > 0x20)
               {
                   if ((i1 + 1) <= 40)
                   {
                       /* Ecrit les espaces au format ASM */
                       i0 = i1 - i0 ;
                       while(i0 != 0)
                       {
                           *(dst++) = (UBYTE)(0xF0 | ((i0 > 15) ? 0x0F : i0)) ;
                           i0 -= (i0 > 15) ? 0x0F : i0 ;
                       }
                       if (c == (UBYTE)CHAR127) c = 0x7F ;
                       if (c >= 0x80)
                           c = (UBYTE)txt_to_asm_t[c - 0x80] ;
                       *(dst++) = (UBYTE)c ;
                       i1++ ;
                       i0 = i1 ;
                   }
               }
        }

        /* Agrandi le buffer si ncessaire */
        if ((size_t)(dst-temp_buffer) > (temp_size - 20))
        {
            dst -= (size_t)temp_buffer ;
            IncreaseTempBuffer(1024) ;
            dst += (size_t)temp_buffer ;
        }
    }
    FreeTemp((size_t)(dst-temp_buffer)) ;

    return TRUE ;
}





static int asm_to_txt_t[16] = {
      0xe7, 0xe1, 0xe2, 0xe4,
      0xe0, 0xe9, 0xea, 0xeb,
      0xe8, 0xee, 0xef, 0xf4,
      0xf6, 0xfb, 0xfc, 0xf9
};


/*
 * Traitement d'un fichier ASM (TO->PC)
 */

int AsmToTxt(void)
{
    int i ;
    int c ;
    UBYTE *src ;
    UBYTE *dst ;

    AllocTemp(1024) ;
    if ((temp_buffer == NULL) || (file_buffer == NULL))
        return PError("", ERR_NOT_ENOUGH_MEMORY) ;
    src = file_buffer ;
    dst = temp_buffer ;

    while ((size_t)(src-file_buffer) < file_size)
    {
        c = (int)*(src++) ;

        if (c == 127) c = (int)CHAR127 ;
        if (c == 0x0D) c = 0x0A ;

        if (c > 0xF0)
        {
            for(i=c;i>0xF1;i--)
                *(dst++) = 0x20 ;
            c = 0x20 ;
        }
        else
        if ((c >= 0x80) && (c <= 0x8F))
            c = asm_to_txt_t[c - 128] ;

        *(dst++) = (UBYTE)c ;

        /* Agrandi le buffer si ncessaire */
        if ((size_t)(dst-temp_buffer) > (temp_size - 20))
        {
            dst -= (size_t)temp_buffer ;
            IncreaseTempBuffer(1024) ;
            dst += (size_t)temp_buffer ;
        }
    }
    FreeTemp((size_t)(dst-temp_buffer)) ;

    return TRUE ;
}
