/*       ___   ___   ___   ___
 *      /_  \ /   \ /  _\ /  _\
 *     // \_/// \_/_\ / \\  / \\
 *    //  _ //  _ / \\\_// /  //
 *    \\_/ \\\_/ \\ //  /  \_//
 *     \___/ \___/ \\__/ \___/
 *     Transferts srie pour Thomson
 *
 *  Conversion de fichiers RAW<->BIN
 *
 *  Copyright (C) 2015 Puls
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <string.h>

#include "defs.h"

extern UBYTE *file_buffer ;
extern ULONG file_size ;
extern UBYTE *temp_buffer ;
extern ULONG temp_size ;
extern void AllocTemp(size_t) ;
extern void FreeTemp(size_t) ;

extern struct DISK_T dk[4] ;

extern int PError(char *, int) ;

/*
 * Traitement d'un fichier BIN (PC->TO)
 * Cette fonction effectue la "binairisation"
 * de n'importe quel fichier
 */
int RawToBin(void)
{
    UBYTE *dst ;

    /* Alloue la mmoire pour la conversion du fichier */
    AllocTemp(file_size + 10) ;
    if ((file_buffer == NULL) || (temp_buffer == NULL))
        return PError("", ERR_NOT_ENOUGH_MEMORY) ;
    dst = temp_buffer ;
    
    /* Limite la taille du fichier */
    if (file_size > 65535)
        return PError("", ERR_BIN_FILE_TOO_LONG) ;

    /* Ecrit l'en-tte de fichier binaire */
    memset(dst, 0, 5);
    *(dst+1) = (UBYTE)(file_size >> 8) ;
    *(dst+2) = (UBYTE)file_size ;

    /* Copie le fichier */
    memcpy(dst+5,file_buffer,file_size) ;

    /* Ecrit la clture de fichier binaire */
    *(dst+5+file_size) = 0xFF ;
    memset(dst+6+file_size , 0, 4);

    FreeTemp(file_size + 10) ;

    return TRUE ;
}



/*
 * Epure un fichier binaire de ses marquages
 */
void ExtractBin(void)
{
    UBYTE *src = file_buffer ;
    size_t size ;
    int again ;

    do
    {
        again = FALSE ;
        if (file_size > 10)
        {
            size = ((size_t)*(src+1) << 8) + (size_t)*(src+2) + 10 ;
            if ((*src == 0x00)
             && (file_size == size + 10)
             && (*(src+5+size) == 0xFF)
             && (*(src+5+size+1) == 0x00)
             && (*(src+5+size+2) == 0x00))
            {
                memmove(src,src+5,size) ;
                file_size -= 10 ;
                again = TRUE ;
            }
        }
    } while (again == TRUE) ;
}



/*
 * Traitement d'un fichier BIN (TO->PC)
 */
int BinToRaw(void)
{
    UBYTE *src = file_buffer ;
    UBYTE *dst = file_buffer ;
    size_t hunksize ;

    /* Ecriture du fichier */
    while (*src == 0x00)
    {
        /* Rcupre la taille du hunk */
        hunksize = ((size_t)*(src+1) << 8) | (size_t)*(src+2) ;
        src += 5 ;
        if (((size_t)src + hunksize - (size_t)file_buffer) >= file_size)
            return PError("", ERR_BAD_FILE_FORMAT) ;

        /* Copie le hunk binaire */
        memcpy(dst, src, hunksize) ;
        dst += hunksize ;
        src += hunksize ;
    }

    /* Officialisation de la nouvelle taille de fichier */
    file_size = (size_t)(dst - file_buffer) ;

    return TRUE ;
}
