/*       ___   ___   ___   ___
 *      /_  \ /   \ /  _\ /  _\
 *     // \_/// \_/_\ / \\  / \\
 *    //  _ //  _ / \\\_// /  //
 *    \\_/ \\\_/ \\ //  /  \_//
 *     \___/ \___/ \\__/ \___/
 *     Transferts srie pour Thomson
 *
 *  Conversion de fichiers MAP/PPM<->BMP
 *
 *  Copyright (C) 2015 Puls
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>  /******************************/
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>

#include "defs.h"

#ifdef DEBUG_MODE
    extern FILE *fp_map ;
#endif

extern ULONG ReadLELong(UBYTE *) ;
extern USHORT ReadLEWord(UBYTE *) ;
extern USHORT ReadBEWord(UBYTE *) ;
extern void WriteLELong(UBYTE *, ULONG) ;
extern void WriteLEWord(UBYTE *, USHORT) ;
extern void WriteBEWord(UBYTE *, USHORT) ;

extern int RawToBin(void) ;

extern int PError(char *, int) ;

extern struct CC90_T cc90 ;
extern struct CONV_T conv ;

extern UBYTE *file_buffer ;
extern ULONG file_size ;
extern UBYTE *temp_buffer ;
extern ULONG temp_size ;
extern void AllocTemp(size_t) ;
extern void FreeTemp(size_t) ;

static int histogram[16] ;       /* Table histogrammique */

/*
 * Table de gamma correction TO <-> PC
 * Composante_PC = ((Composante_TO/15)^(1/2.515))*255
 */
static ULONG mapgtab[16] = {
    0x00, 0x56, 0x72, 0x86,
    0x96, 0xa4, 0xb1, 0xbc,
    0xc6, 0xd0, 0xd9, 0xe1,
    0xe9, 0xf0, 0xf8, 0xff
};

static struct {
/*  USHORT bfType ;        /  0 */
    ULONG  bfSize ;        /* 2 */
/*  USHORT bfReserved1 ;   /  6 */
/*  USHORT bfReserved2 ;   /  8 */
    ULONG  bfOffBits ;     /* 10 */
    ULONG  biSize ;        /* 14 */
    ULONG  biWidth ;       /* 18 */
    ULONG  biHeight ;      /* 22 */
/*  USHORT biPlanes ;      /  26 */
    USHORT biBitCount ;    /* 28 */
/*  ULONG  biCompression ; /  30 */
    /* Dans un fichier BMP, les informations qui
     * suivent ne sont pas toujours mises  jour
     * Il est alors prfrable de calculer la
     * taille du bitmap  partir des dimensions
     * de l'image, et mettre le nombre de couleurs
     * au maximum quand il est  0 */
    ULONG  biSizeImage ;     /* 34 */
/*  ULONG  biXPelsPerMeter ; /  38 */
/*  ULONG  biYPelsPerMeter ; /  42 */
    ULONG  biClrUsed ;       /* 46 */
    ULONG  biClrImportant ;  /* 50 */
} bmp ;
#define BITMAPHEADERSIZE  54

/* Structure pour fichiers MAP */
static struct {
        ULONG  width ;
        ULONG  height ;
        USHORT scrmod ;
        USHORT video ;
        USHORT outcol ;
        int    usedcol ;
} map ;

enum {
    BYTE_FREE,
    BYTE_EQUAL,
    BYTE_NOT_EQUAL,
    BYTE_STOP
} ;

#define   INDEX_STOP    256
#define   INDEX_END     (INDEX_STOP * 2) + 1
#define   SEQ_MAX_SIZE  255

static struct {
    int    type ;
    size_t size ;
    int    next ;
} seq[INDEX_END] ;


/*
 * Recherche une squence avantageuse
 */
static void FindMapSeq(int i, int *gain, int *number)
{
    int isize, count=0, icount=0 ;
    int size ;
    int gain0, number0 = 0 ;

    *number = 1 ;
    *gain = -3 ;

    while(i != INDEX_STOP)
    {
        /* Engage la taille de la squence */
        size = (int)seq[i].size % SEQ_MAX_SIZE ;
        count += ((seq[i].type == BYTE_EQUAL) ? ((size != 0) ? 2 : 0) : ((size != 0) ? 2 : 0) + size) ;
        icount += size ;
        isize = (((icount + SEQ_MAX_SIZE - 1) / SEQ_MAX_SIZE) * 2) + icount ;
        number0++ ;

        /* Si gain suffisant   : reprage
         * Si gain insuffisant : interruption */
        if (number0 > 1)
        {
            /* Calcule le gain */
            gain0 = count - isize ;

            /* Retour si gain insuffisant */
            if ((gain0 - *gain) < - 1) return ;

            /* Reprage d'un gain avantageux */
            if ((((float)gain0 / (float)number0) - ((float)*gain / (float)*number)) > FLT_EPSILON)
            {
                *gain = gain0 ;
                *number = number0 ;
            }

            /* Complte la taille de la squence */            
            size = (int)seq[i].size / SEQ_MAX_SIZE ;
            count += (seq[i].type == BYTE_EQUAL) ? 2 * size : (2 + SEQ_MAX_SIZE) * size ;
            icount += size * SEQ_MAX_SIZE ;
        }
        i = seq[i].next ;
    }
}


/*
 * Compression d'un plan de type MAP
 * Retour = taille de la compression
 */
static int
CrunchMapPlane(UBYTE *src, UBYTE *dst, size_t size)
{
    int i, i0 ;
    int found ;
    UBYTE c ;
    UBYTE *ptr ;
    int first_element = 0, last_element = 0 ;
    size_t seqsize ;
    int number = 0, number0 = 1 ;
    int gain = 0, gain0 = 0 ;
    int planesize = 0 ;

    /* Efface la structure des squences */
    memset(&seq,0x00,sizeof(seq)) ;

    /* Le BYTE_STOP (terminateur de liste) boucle sur lui-mme */
    seq[INDEX_STOP].type = BYTE_STOP ;
    seq[INDEX_STOP].size = (size_t)INT_FAST16_MAX ; /* Valeur grande et arbitraire */
    seq[INDEX_STOP].next = INDEX_STOP ;

    ptr = src ;

    while (size != 0)
    {
        /* -----------------------------
         * Cre la liste des squences
         */
        i = 0 ;
        while((size != 0) && (i < INDEX_STOP))
        {
            /* Recherche une place libre */
            while ((seq[i].type != BYTE_FREE) && (i < INDEX_STOP)) i++ ;
            
            /* Ajoute l'lment si place libre trouve */
            if (i < INDEX_STOP)
            {
                seqsize = 0 ;
                c = *ptr ;
                if (*(ptr+1) == c)
                {
                    seq[i].type = BYTE_EQUAL ;
                    while(((size - seqsize) != 0) && (ptr[seqsize] == c))
                        seqsize++ ;
                }
                else
                {
                    seq[i].type = BYTE_NOT_EQUAL ;
                    while(((size - seqsize) != 0) && (ptr[seqsize] != ptr[seqsize+1]))
                        seqsize++ ;
                }
                /* Passe les compteurs  1 en galit */   
                if (seqsize == 1)
                    seq[i].type = BYTE_EQUAL ;
                
                seq[i].size = seqsize ;
                size -= seqsize ;
                ptr += seqsize ;

                seq[last_element].next = i ;
                last_element = i ;
            }
        }
        /* Lie le dernier lment de la liste au terminateur */
        seq[last_element].next = INDEX_STOP ;

        /* -------------------------
         * Compresse les squences
         */

        /* Inclusion si possible en squence d'ingalit */
        found = TRUE ;
        while(found == TRUE)
        {
            found = FALSE ;
            gain0 = 0 ;
            number0 = 2 ;
            i0 = 0 ;
            i = first_element ;
            while(i != INDEX_STOP)
            {
                FindMapSeq(i,&gain,&number) ;
                if ((((float)gain / (float)number) - ((float)gain0 / (float)number0)) > FLT_EPSILON)
                {
                    number0 = number ;
                    gain0 = gain ;
                    i0 = i ;
                    found = TRUE ;
                }
                i = seq[i].next ;
            }
            if (found == TRUE)
            {
                /* Cas particulier d'une 1re squence en galit */
                if ((seq[i0].type == BYTE_EQUAL)
                 && (seq[i0].size > SEQ_MAX_SIZE))
                {
                    i = 0 ;
                    while ((seq[i].type != BYTE_FREE) && (i < INDEX_END)) i++ ;
                    seq[i].next = seq[i0].next ;
                    seq[i0].next = i ;
                    seq[i].size = seq[i0].size % SEQ_MAX_SIZE ;
                    seq[i0].size -=  seq[i].size ;
                    i0 = i ;
                }

                /* Fusion des 2 premiers lments */
                i = i0 ;
                seq[i].type = BYTE_NOT_EQUAL ;
                i0 = seq[i].next ;
                seq[i].size += seq[i0].size % SEQ_MAX_SIZE ;
                
                /* Fusion des lments intermdiaires */
                while(number0 > 2)
                {
                    seq[i].size += seq[i0].size - (seq[i0].size % SEQ_MAX_SIZE) ;
                    seq[i0].type = BYTE_FREE ;
                    i0 = seq[i0].next ;
                    seq[i].size += seq[i0].size % SEQ_MAX_SIZE ;
                    number0-- ;
                }
                /* Fusion du dernier lment */
                seq[i0].size -= seq[i0].size % SEQ_MAX_SIZE ;
                if ((seq[i0].size >= SEQ_MAX_SIZE)
                 && (seq[i0].type == BYTE_EQUAL))
                    seq[i].next = i0 ;
                else
                {
                    seq[i].next = seq[i0].next ;
                    seq[i].size += seq[i0].size ;
                    seq[i0].type = BYTE_FREE ;
                }
            }
        }


        /* -----------------------------------------
         * Repre le dernier lment si pas fin
         * Il est prfrable, pour une bonne compression,
         * de conserver la dernire ingalit de la liste
         * ou au moins la dernire galit dont le compteur
         * est le plus grand possible
         */
        last_element = INDEX_STOP ;
        if (size != 0)
        {
            i = first_element ;
            while (i != INDEX_STOP)
            {
                if ((seq[i].type == BYTE_NOT_EQUAL)
                 || ((seq[i].type == BYTE_EQUAL) && (seq[i].size > 4)))
                    last_element = i ;
                i = seq[i].next ;
            }
        }

        /* -----------------------
         * Ecrit la compression
         */

        i = first_element ;
        while(i != last_element)
        {
            if (seq[i].type == BYTE_EQUAL)
            {
                while(seq[i].size != 0)
                {
                    seqsize = (seq[i].size > SEQ_MAX_SIZE) ? (size_t)SEQ_MAX_SIZE : seq[i].size ;
                    planesize += 2 ;
                    seq[i].size -= seqsize ;
                    *(dst++) = (UBYTE)seqsize ;
                    *(dst++) = *src ;
#ifdef DEBUG_MODE
                    fprintf(fp_map,"e%d(%02x) ",seqsize,*src) ;
#endif
                    src += seqsize ;
                }
            }
            else
            if (seq[i].type == BYTE_NOT_EQUAL)
            {
                while(seq[i].size != 0)
                {
                    *(dst++) = 0x00 ;
                    seqsize = (seq[i].size > SEQ_MAX_SIZE) ? (size_t)SEQ_MAX_SIZE : seq[i].size ;
                    planesize += 2 + seqsize ;
                    seq[i].size -= seqsize ;
                    *(dst++) = (UBYTE)seqsize ;
#ifdef DEBUG_MODE
                    fprintf(fp_map,"i(%d)[",seqsize) ;
#endif
                    while(seqsize != 0)
                    {
                        *(dst++) = *src ;
#ifdef DEBUG_MODE
                        fprintf(fp_map,"%02x ",*src) ;
#endif
                        src++ ;
                        seqsize-- ;
                    }
#ifdef DEBUG_MODE
                    fprintf(fp_map,"]\n") ;
#endif
                }
            }
            seq[i].type = BYTE_FREE ;
            i = seq[i].next ;
        }
        /* Repre le premier lment */
        first_element = last_element ;
        i = last_element ;
        /* Repre le dernier lment */
        while(i != INDEX_STOP)
        {
            last_element = i ;
            i = seq[i].next ;
        }
    }
    return planesize ;
}



/*
 * Cre l'histogramme du graphique
 */
static void
CreateHistogram(UBYTE *bitmap)
{
    ULONG x , y ;
    UBYTE mask ;
    UBYTE shift ;
    UBYTE *ptr_init = bitmap ;
    UBYTE *ptr ;

    memset(histogram,0x00,sizeof(histogram)) ;

    for(y=0;y<bmp.biHeight;y++)
    {
        shift = 8 ;
        ptr = ptr_init ;
        mask = (UBYTE)(0xFF >> (8 - bmp.biBitCount)) ;
        for(x=0;x<bmp.biWidth;x++)
        {
            shift -= bmp.biBitCount ;
            histogram[(int)((~*ptr >> shift) & mask)]++ ;
            if (shift == 0) { ptr++ ; shift = 8 ; }
        }
        ptr_init += (int)(bmp.biSizeImage / bmp.biHeight) ;
    }
}


/*
 * Reprage de la couleur de tour pour TO-SNAP
 *
 * La couleur de tour d'cran est dtermine par la couleur la plus
 * frquente au pourtour intrieur de l'cran pour tout type
 */
static UBYTE
GetMapBackGround(UBYTE *bitmap)
{
    int i ;
    UBYTE color ;
    ULONG count_table[16] ;
    UBYTE mask = (UBYTE)((bmp.biBitCount == 1) ? 0x01 : 0x0F) ;
    ULONG hard_width = bmp.biSizeImage / bmp.biHeight ;
    ULONG shift ;
    UBYTE *ptr ;

    memset(count_table,0x00,sizeof(count_table)) ;
    shift = 8 ;
    ptr = bitmap ;

    /* Compte pour ligne haute */
    for(i=0;i<(int)bmp.biWidth;i++) {
         if (shift == 0) { ptr++ ; shift = 8 ; }
         shift -= bmp.biBitCount ;
         count_table[((ULONG)~*ptr >> shift) & mask]++ ;
    }
    /* Compte pour colonne droite */
    for(i=1;i<(int)bmp.biHeight;i++) {
         ptr += hard_width ;
         count_table[(((ULONG)~*ptr & 0xFF) >> shift) & mask]++ ;
    }
    /* Compte pour ligne basse */
    for(i=1;i<(int)bmp.biWidth;i++) {
         shift += bmp.biBitCount ;
         if (shift == 8) { ptr-- ; shift = 0 ; }
         count_table[((ULONG)~*ptr >> shift) & mask]++ ;
    }
    /* Compte pour colonne gauche */
    for(i=1;i<(int)bmp.biHeight-1;i++) {
         ptr -= hard_width ;
         count_table[((ULONG)~*ptr >> shift) & mask]++ ;
    }
    /* Reprage de la couleur la plus frquente */
    color = 0 ;
    for (i=0;i<16;i++)
        if (count_table[i] > count_table[color])
            color = (UBYTE)i ;

    return color ;
}


/*
 * Unifie le bitplan
 *
 * Les couleurs les plus loignes dans la palette
 * sont ventuellement remplaces par des couleurs
 * quivalentes plus proches du dbut de la palette
 */
static void
UnifyBitPlane(USHORT *palette, UBYTE *bitmap)
{
    ULONG x,y ;
    int c ;
    UBYTE shift ;
    UBYTE corresp[16] ;
    UBYTE mask ;
    UBYTE *ptr_init = bitmap ;
    UBYTE *ptr ;

    for(x=0;x<16;x++)
    {
        y = 0 ;
        while(palette[x] != palette[y]) y++ ;
        corresp[x] = y ;
    }

    for(y=0;y<bmp.biHeight;y++)
    {
        shift = 8 ;
        ptr = ptr_init ;
        mask = (UBYTE)((bmp.biBitCount == 1) ? 0x01 : ((map.usedcol == 16) ? 0x0F : 0x03)) ;
        for(x=0;x<bmp.biWidth;x++)
        {
            shift -= bmp.biBitCount ;
            c = (int)((ULONG)~*ptr >> shift) & mask ;
            *ptr = ~((~*ptr & (~(mask << shift) & 0xFF)) | (corresp[c] << shift)) & 0xFF ; 
            if (shift == 0) { ptr++ ; shift = 8 ; }
        }
        ptr_init += (int)(bmp.biSizeImage / bmp.biHeight) ;
    }
}


/*--------------------------------------------------------------------------
 * Traitement d'un fichier BMP (PC->TO)
 */


/*
 * Remerciements  Samuel Devulder pour son aide prcieuse
 */
static float
dist(float *t1, float *t2)
{
   float b1 = t1[0] ;
   float v1 = t1[1] ;
   float r1 = t1[2] ;
   float b2 = t2[0] ;
   float v2 = t2[1] ;
   float r2 = t2[2] ;

   b1 -= b2 ;
   v1 -= v2 ;
   r1 -= r2 ;

   return ((b1*b1) + (v1*v1) + (r1*r1)) ;
}


/*
 * Dither sur une ligne et retourne l'erreur sur l'octet
 * Le programme en profite pour calculer l'octet de forme
 * Remerciements  Samuel Devulder pour son aide prcieuse
 */
static int
line_dith_error(USHORT *cmap, UBYTE *mem, UBYTE *res_forme, UBYTE forme_c, UBYTE fond_c)
{
    ULONG i ;
    int byte ;
    float error = 0 ;
    float d1, d2 ;
    float forme[3], fond[3] ;
    float add[3], color[3]  ;

    forme[0] = (float)(cmap[forme_c] & 0x0F) ;
    forme[1] = (float)((cmap[forme_c] >> 4) & 0x0F) ;
    forme[2] = (float)((cmap[forme_c] >> 8) & 0x0F) ;
    fond[0] = (float)(cmap[fond_c] & 0x0F) ;
    fond[1] = (float)((cmap[fond_c] >> 4) & 0x0F) ;
    fond[2] = (float)((cmap[fond_c] >> 8) & 0x0F) ;
    add[0] = 0 ;
    add[1] = 0 ;
    add[2] = 0 ;

    *res_forme = 0x00 ;
    for(i=0;i<8;i++)
    {
        byte = (int)((~mem[i >> 1] >> ((~i & 1) * 4)) & 0x0F) ;
        color[0] = (float)(cmap[byte] & 0x0F) + add[0] ;
        color[1] = (float)((cmap[byte] >> 4) & 0x0F) + add[1] ;
        color[2] = (float)((cmap[byte] >> 8) & 0x0F) + add[2] ;
        d1 = dist(color,forme) ;
        d2 = dist(color,fond) ;
        error += ((d2 - d1) > FLT_EPSILON) ? d1 : d2 ;
        *res_forme <<= 1 ;
        *res_forme |= ((d2 - d1) > FLT_EPSILON) ? 0 : 1 ; 
        add[0] = (((((d2 - d1) > FLT_EPSILON) ? forme[0] : fond[0]) - color[0]) * 7) / 16 ;
        add[1] = (((((d2 - d1) > FLT_EPSILON) ? forme[1] : fond[1]) - color[1]) * 7) / 16 ;
        add[2] = (((((d2 - d1) > FLT_EPSILON) ? forme[2] : fond[2]) - color[2]) * 7) / 16 ;
    }
    return (int)error ;
}


/*
 * Reprage du meilleur couple forme/fond
 * Remerciements  Samuel Devulder pour son aide prcieuse
 */
static void
get_bytes_40(USHORT *cmap, UBYTE *mem, UBYTE *res_forme, UBYTE *res_fond)
{
    ULONG i, j ;
    int error ;
    int errmem = INT32_MAX ;
    UBYTE forme = 0 ;
    ULONG nbcolors = 0 ;
    UBYTE color_byte[8] ;
    UBYTE color_list[8] ;
    int color_freq[16] ;
    UBYTE digit = 0, foreg = 0, backg = 0 ;

    memset(color_freq,0x00,sizeof(color_freq)) ;
    memset(color_byte,0x00,sizeof(color_byte)) ;
    memset(color_list,0x00,sizeof(color_list)) ;

    for(i=0;i<8;i++)
    {
        color_byte[i] =(UBYTE)((~mem[i >> 1] >> ((~i & 1) * 4)) & 0x0F) ;
        if (color_freq[color_byte[i]]++ == 0)
            color_list[nbcolors++] = color_byte[i] ;
    } 

    switch(nbcolors)
    {
        case 1 :
            digit = 0x00 ;
            foreg = color_list[0] ;
            backg = color_list[0] ;
            break ;
     
        case 2 :
            for(i=0;i<8;i++)
            {
                digit <<= 1 ;
                digit |= (color_byte[i] == color_list[0]) ? 0 : 1 ;
            }
            backg = color_list[0] ;
            foreg = color_list[1] ;
            break ;

        default :
        /*
            for(i=0;i<16;i++)
            {
                for(j=i+1;j<16;j++)
                {
                    error = line_dith_error(cmap,mem,&forme,i,j) ;
                    if (error < errmem)
                    {
                        errmem = error ;
                        b.dig[1] = forme ;
                        b.frg[1] = j ;
                        b.bkg[1] = i ;
                    }
                }
            }
            */
            for(i=0;i<nbcolors;i++)
            {
                for(j=0;j<nbcolors;j++)
                {
                    error = line_dith_error(cmap,mem,&forme,color_list[i],color_list[j]) ;
                    if (error < errmem)
                    {
                        errmem = error ;
                        digit = forme ;
                        foreg = color_list[j] ;
                        backg = color_list[i] ;
                    }
                }
            }
            break ;
   }

    *res_forme = (UBYTE)digit ;
    *res_fond = (UBYTE)((foreg << 3) | ((backg + 0x78) & 0x87)) ^ 0xC0 ;
}



enum {
    PREV_SEQ,
    CURR_SEQ
} ;

enum {
    MEM0_PTR,
    MEM1_PTR,
    MEM2_PTR
} ;

struct BYTE_STRUCT {
    UBYTE  dig[CURR_SEQ+1] ;
    UBYTE  frg[CURR_SEQ+1] ;
    UBYTE  bkg[CURR_SEQ+1] ;
    UBYTE  *ptr[MEM2_PTR+1] ;
} ;


/*
 * Capture digit
 */
static void
GetDigit40(UBYTE *bitmap, ULONG planesize, struct BYTE_STRUCT *b)
{
    b->dig[CURR_SEQ] = *bitmap ;
    b->frg[CURR_SEQ] = ((*(bitmap + planesize) >> 3) & 0x0F) ^ 0x08 ;
    b->bkg[CURR_SEQ] = (*(bitmap + planesize) & 0x87) ;
    b->bkg[CURR_SEQ] = ((b->bkg[CURR_SEQ] & 0x80) == 0)
                             ? (b->bkg[CURR_SEQ] & 0x07) | 0x08
                             : b->bkg[CURR_SEQ] & 0x07 ;
}


/*
 * Swappe et actualise couleurs si correspondent
 */
static void
SwapColors40(UBYTE *bitmap, ULONG planesize, struct BYTE_STRUCT *b)
{
    if ((b->bkg[CURR_SEQ] == b->frg[PREV_SEQ])
     && (b->bkg[PREV_SEQ] == b->frg[CURR_SEQ]))
    {
        b->dig[CURR_SEQ] = (~b->dig[CURR_SEQ] & 0xFF) ;
        b->frg[CURR_SEQ] = b->frg[PREV_SEQ] ;
        b->bkg[CURR_SEQ] = b->bkg[PREV_SEQ] ;
        *bitmap = (UBYTE)b->dig[CURR_SEQ] ;
        *(bitmap + planesize) = (UBYTE)((b->frg[CURR_SEQ] << 3) 
                             | ((b->bkg[CURR_SEQ] + 0x78) & 0x87)) ^ 0xC0 ;
    }
}


/*
 * Ajuste les couleurs d'une squence "nulle"
 */
static void
AjustColors40(UBYTE *bitmap, UBYTE *bitmap_end, ULONG planesize, struct BYTE_STRUCT *b)
{
    UBYTE *ptr ;
    int i = ((b->ptr[MEM1_PTR]-b->ptr[MEM0_PTR]) > (bitmap-b->ptr[MEM2_PTR])) ? MEM0_PTR : MEM2_PTR ;
    UBYTE frg, bkg ;

    GetDigit40(b->ptr[i],planesize,b) ;
    frg = b->frg[CURR_SEQ] ;
    bkg = b->bkg[CURR_SEQ] ;
    GetDigit40(b->ptr[MEM1_PTR],planesize,b) ;
    if ((b->dig[CURR_SEQ] == 0x00)
     && (b->bkg[CURR_SEQ] != bkg)
     && (b->bkg[CURR_SEQ] != frg))
        i = (i == MEM0_PTR) ? MEM2_PTR : MEM0_PTR ;
    else
    if ((b->dig[CURR_SEQ] == 0xFF)
     && (b->frg[CURR_SEQ] != bkg)
     && (b->frg[CURR_SEQ] != frg))
        i = (i == MEM0_PTR) ? MEM2_PTR : MEM0_PTR ;
    GetDigit40(b->ptr[i],planesize,b) ;
    frg = b->frg[CURR_SEQ] ;
    bkg = b->bkg[CURR_SEQ] ;

#ifdef DEBUG_MODE
    fprintf(fp_map,"<actualize ") ;
#endif

    for(ptr=b->ptr[MEM1_PTR];ptr<b->ptr[MEM2_PTR];ptr++)
    {
        GetDigit40(ptr,planesize,b) ;
        if (b->dig[CURR_SEQ] == 0x00)
        {
            if (b->bkg[CURR_SEQ] == bkg)
                b->frg[CURR_SEQ] = frg ;
            else
            {
                b->dig[CURR_SEQ] = 0xFF ;
                b->frg[CURR_SEQ] = frg ;
                b->bkg[CURR_SEQ] = bkg ;
            }
        }
        else
        if (b->dig[CURR_SEQ] == 0xFF)
        {
            if (b->frg[CURR_SEQ] == frg)
                b->bkg[CURR_SEQ] = bkg ;
            else
            {
                b->dig[CURR_SEQ] = 0x00 ;
                b->frg[CURR_SEQ] = frg ;
                b->bkg[CURR_SEQ] = bkg ;
            }
        }
        *ptr = (UBYTE)b->dig[CURR_SEQ] ;
        *(ptr + planesize) = (UBYTE)((b->frg[CURR_SEQ] << 3)
                             | ((b->bkg[CURR_SEQ] + 0x78) & 0x87)) ^ 0xC0 ;
#ifdef DEBUG_MODE
        fprintf(fp_map,"%02x:%01x%01x ",b->dig[CURR_SEQ],b->frg[CURR_SEQ],b->bkg[CURR_SEQ]) ;
#endif
    }
#ifdef DEBUG_MODE
    fprintf(fp_map,">\n") ;
#endif
    /* Rinitialise les paramtres courants */
    if (bitmap < bitmap_end)
        GetDigit40(bitmap,planesize,b) ;
}


/*
 * Optimisation de la compression pour le 40 colonnes
 */
static void
OptimizeCmp40(UBYTE *bitmap, ULONG planesize)
{
    struct BYTE_STRUCT b ;
    UBYTE *bitmap_end ;
    UBYTE dbig = 0 ;
    UBYTE dfrg = 0 ;
    UBYTE dbkg = 0 ;

    bitmap_end = bitmap + planesize ;

    memset(&b,0x00,sizeof(struct BYTE_STRUCT)) ;
    b.ptr[MEM0_PTR] = bitmap ;
    b.ptr[MEM1_PTR] = bitmap ;
    b.ptr[MEM2_PTR] = bitmap ;
    GetDigit40(bitmap,planesize,&b) ;
    b.dig[PREV_SEQ] = b.dig[CURR_SEQ] ;
    b.frg[PREV_SEQ] = b.frg[CURR_SEQ] ;
    b.bkg[PREV_SEQ] = b.bkg[CURR_SEQ] ;

    while(bitmap < bitmap_end)
    {
        /*----------------------------------------------------------------------*/
        /* Passe la squence de dpart "non nulle" (digit!=0x00 ou digit!=0xFF) */
        while(((b.dig[CURR_SEQ] != 0x00) && (b.dig[CURR_SEQ] != 0xFF))
          && (bitmap < bitmap_end))
        {        
#ifdef DEBUG_MODE
            fprintf(fp_map,"%02x:%01x%01x ",b.dig[CURR_SEQ],b.frg[CURR_SEQ],b.bkg[CURR_SEQ]) ;
#endif
            SwapColors40(bitmap,planesize,&b) ;
            /* Actualise pointeur si couleurs diffrentes */
            if ((b.frg[CURR_SEQ] != b.frg[PREV_SEQ])
             || (b.bkg[CURR_SEQ] != b.bkg[PREV_SEQ]))
            {
                b.ptr[MEM0_PTR] = bitmap ;
#ifdef DEBUG_MODE
                fprintf(fp_map,"<new start> ") ;
#endif
            }
            b.dig[PREV_SEQ] = b.dig[CURR_SEQ] ;
            b.frg[PREV_SEQ] = b.frg[CURR_SEQ] ;
            b.bkg[PREV_SEQ] = b.bkg[CURR_SEQ] ;
            /* Rcupre digit suivant */
#ifdef DEBUG_MODE
            fprintf(fp_map,">> %02x:%01x%01x\n",b.dig[CURR_SEQ],b.frg[CURR_SEQ],b.bkg[CURR_SEQ]) ;
#endif
            bitmap++ ;
            if (bitmap < bitmap_end)
                GetDigit40(bitmap,planesize,&b) ;
        }

        /*------------------------------------------------------*/
        /* Passe la squence "nulle" (digit=0x00 ou digit=0xFF) */
        b.ptr[MEM1_PTR] = bitmap  ;
        b.ptr[MEM2_PTR] = bitmap ;
        while(((b.dig[CURR_SEQ] == 0x00) || (b.dig[CURR_SEQ] == 0xFF))
          && (bitmap < bitmap_end))
        {
            b.ptr[MEM2_PTR] = bitmap ;
            dbig = b.dig[CURR_SEQ] ;
            dfrg = b.frg[CURR_SEQ] ;
            dbkg = b.bkg[CURR_SEQ] ;

#ifdef DEBUG_MODE
            fprintf(fp_map,"%02x:%01x%01x ",b.dig[CURR_SEQ],b.frg[CURR_SEQ],b.bkg[CURR_SEQ]) ;
#endif
            if (((b.dig[CURR_SEQ] == 0x00)
              && (b.bkg[CURR_SEQ] != b.bkg[PREV_SEQ])
              && (b.bkg[CURR_SEQ] != b.frg[PREV_SEQ]))
             || ((b.dig[CURR_SEQ] == 0xFF)
              && (b.frg[CURR_SEQ] != b.bkg[PREV_SEQ])
              && (b.frg[CURR_SEQ] != b.frg[PREV_SEQ])))
            {
                /* Ajuste la squence */
                AjustColors40(bitmap,bitmap_end,planesize,&b) ;
                b.ptr[MEM0_PTR] = bitmap ;
                b.ptr[MEM1_PTR] = bitmap ;
                b.ptr[MEM2_PTR] = bitmap ;
                b.dig[PREV_SEQ] = b.dig[CURR_SEQ] ;
                b.frg[PREV_SEQ] = b.frg[CURR_SEQ] ;
                b.bkg[PREV_SEQ] = b.bkg[CURR_SEQ] ;
                dbig = b.dig[CURR_SEQ] ;
                dfrg = b.frg[CURR_SEQ] ;
                dbkg = b.bkg[CURR_SEQ] ;
            }
#ifdef DEBUG_MODE
            fprintf(fp_map,"\n") ;
#endif
            bitmap++ ;
            if (bitmap < bitmap_end)
                GetDigit40(bitmap,planesize,&b) ;
        }

        /*-------------------------------------------------------------------*/
        /* Passe la squence de fin "non nulle" (digit!=0x00 ou digit!=0xFF) */
        b.ptr[MEM2_PTR] = bitmap ;
        SwapColors40(bitmap,planesize,&b) ;
#ifdef DEBUG_MODE
        fprintf(fp_map,"<start next> ") ;
#endif
        if ((b.ptr[MEM1_PTR] != b.ptr[MEM2_PTR])
           && (((dbig == 0x00)
             && ((dbkg == b.bkg[CURR_SEQ]) || (dbkg == b.frg[CURR_SEQ])))
            || ((dbig == 0xFF)
             && ((dfrg == b.bkg[CURR_SEQ]) || (dfrg == b.frg[CURR_SEQ])))))
        {
#ifdef DEBUG_MODE
            fprintf(fp_map,"<search>\n") ;
#endif
            b.dig[PREV_SEQ] = b.dig[CURR_SEQ] ;
            b.frg[PREV_SEQ] = b.frg[CURR_SEQ] ;
            b.bkg[PREV_SEQ] = b.bkg[CURR_SEQ] ;
            while(((b.dig[CURR_SEQ] != 0x00) && (b.dig[CURR_SEQ] != 0xFF))
                && (((b.frg[CURR_SEQ] == b.frg[PREV_SEQ]) && (b.bkg[CURR_SEQ] == b.bkg[PREV_SEQ]))
                 || ((b.frg[CURR_SEQ] == b.bkg[PREV_SEQ]) && (b.bkg[CURR_SEQ] == b.frg[PREV_SEQ])))
                && (bitmap < bitmap_end))
            {
#ifdef DEBUG_MODE
                fprintf(fp_map,"%02x:%01x%01x\n",b.dig[CURR_SEQ],b.frg[CURR_SEQ],b.bkg[CURR_SEQ]) ;
#endif
                SwapColors40(bitmap,planesize,&b) ;
                b.dig[PREV_SEQ] = b.dig[CURR_SEQ] ;
                b.frg[PREV_SEQ] = b.frg[CURR_SEQ] ;
                b.bkg[PREV_SEQ] = b.bkg[CURR_SEQ] ;
                bitmap++ ;
                if (bitmap < bitmap_end)
                {
                    GetDigit40(bitmap,planesize,&b) ;
                    if ((b.dig[CURR_SEQ] != 0x00) && (b.dig[CURR_SEQ] != 0xFF))
                    {
                        SwapColors40(bitmap,planesize,&b) ;
                    }
                }
            }
        }
        /* Ajuste la squence */
        AjustColors40(bitmap,bitmap_end,planesize,&b) ;
        b.ptr[MEM0_PTR] = ((b.dig[CURR_SEQ] == 0x00) || (b.dig[CURR_SEQ] == 0xFF))
                          ? b.ptr[MEM2_PTR]
                          : bitmap ;
    }
}


/* ===========================================================================
 * Conversion BMP vers MAP/PPM
 * =========================================================================== */


static int
BmpToMap(int maptype)
{
    /* Registres locaux */
    ULONG /*@null@*/ i, j ;
    UBYTE c0 = 0, c1 = 1 ;
    ULONG planesize ;

    ULONG offset ;
    ULONG width ;

    UBYTE *src ;
    UBYTE *dst;
    UBYTE *ptr ;

    USHORT palette[16] ;

    /* Table de gamma correction TO <-> PC */
    UBYTE bmpgtab[256] ;
    
    if (file_buffer == NULL)
        return PError("",ERR_NOT_ENOUGH_MEMORY) ;
    src = file_buffer ;

    /* Vrifie et filtre les infos BMP */
    if ((file_size < BITMAPHEADERSIZE)
     || (ReadLEWord(src+0) != 0x4d42)
     || (ReadLELong(src+18) == 0)     /* bmp.biWidth */
     || (ReadLELong(src+18) > 2040)   /* bmp.biWidth */
     || (ReadLELong(src+22) == 0)     /* bmp.biHeight */
     || (ReadLELong(src+22) > 2040)   /* bmp.biHeight */
     || (ReadLELong(src+50) > ReadLELong(src+46))) /* bmp.biClrImportant > bmp.biClrUsed */
        return PError("",ERR_BAD_FILE_FORMAT) ;
    bmp.biWidth        = ReadLELong(src+18) ;
    bmp.biHeight       = ReadLELong(src+22) ;
    bmp.biBitCount     = ReadLEWord(src+28) ;
    bmp.biClrImportant = ReadLELong(src+50) ;
    bmp.biClrUsed      = ReadLELong(src+46) ;
    if ((ReadLEWord(src+26) != 1)     /* bmp.biPlanes */
     || ((bmp.biBitCount != 4) && (bmp.biBitCount != 1))
     || (bmp.biClrUsed > 16))
        return PError("",ERR_RESOLUTION) ;
    if (ReadLELong(src+30) != 0)      /* bmp.biCompression */
        return PError("",ERR_COMPRESSION) ;
    /* L'ajustement de la taille de la bitmap et du nombre
     * de couleurs est effectu vu que biSizeImage, biClrUsed
     * et biClrImportant ne sont pas ncessairement dfinis
     * dans le fichier BMP */
    bmp.bfSize         = ReadLELong(src+2) ;
    bmp.bfOffBits      = ReadLELong(src+10) ;
    bmp.biSizeImage    = ((((bmp.biWidth * bmp.biBitCount) >> 3) + 3) & 0xFFFFFFFC) * bmp.biHeight ;
    if (bmp.biClrUsed == 0) bmp.biClrUsed = (bmp.bfOffBits - BITMAPHEADERSIZE) >> 2 ;
    if (bmp.biClrImportant == 0) bmp.biClrImportant = bmp.biClrUsed ;

    /* Cration de la table de gamma correction */
    memset(bmpgtab,0x0F,sizeof(bmpgtab)) ;
    j = 0 ;
    for(i=0;i<15;i++)
        while(j<mapgtab[i+1])
            bmpgtab[j++] = i ;

    /* Fixe le mode vido pour le graphique Thomson
     *
     * biBitCount doit tre  4 pour un bitmap 16, un bitmap 4 et
     * un 40 colonnes, ou  1 pour un 80 colonnes
     * Le bitmap 4 est forc si biClrImportant est infrieur ou
     * gal  4
     * Si la video est un bitmap 16, le 40 colonnes est forc si
     * cela est demand ou si la largeur est suprieure  160
     * points */

    switch(bmp.biBitCount)
    {
        case 4  :
            map.video = (USHORT)((bmp.biClrImportant <= 4) ? 2 : 3) ;
            if (map.video == 3)
                if ((conv.force40cols == TRUE) || (bmp.biWidth > 160)) map.video = 0 ;
            break ;
        case 1  :
            map.video = 1 ;
            break ;
    }
    map.width   = (map.video == 3) ? (bmp.biWidth + 1) >> 1 : (bmp.biWidth + 7) >> 3 ;
    map.height  = (bmp.biHeight + 7) >> 3 ;
    map.scrmod  = (USHORT)((map.video & 1) << (7 - (map.video >> 1))) ;
    map.usedcol = (int)(((map.video == 0) || (map.video == 3)) ? 16 : map.video * 2) ;

    /* Limite les dimensions du MAP (doivent contenir sur 8 bits) */
    if ((map.width > 0xFF) || (map.height > 0xFF))
        return PError("",ERR_RESOLUTION) ;

    /* Cre et corrige ventuellement la table de couleur MAP
     * Cette table est faite tt en prvision de la rorganisation des
     * bitplans pour le 40 colonnes. */
    memset(palette, 0, sizeof(palette)) ;
    src = file_buffer + BITMAPHEADERSIZE ;
    for(i=0;i<bmp.biClrImportant;i++)
    {
        /* Traitement composante rouge */
        j = (ReadLELong(src + (i * 4)) >> 16) & 0xFF ;
        palette[map.usedcol - 1 - i] = (conv.gammac != 0) 
                                     ? (USHORT)bmpgtab[j]
                                     : (USHORT)(j >> 4) ;
        /* Traitement composante verte */
        j = (ReadLELong(src + (i * 4)) >> 8) & 0xFF ;
        palette[map.usedcol - 1 - i] |= (conv.gammac != 0)
                                      ? (USHORT)bmpgtab[j] << 4
                                      : (USHORT)j & 0xF0 ;
        /* Traitement composante bleue */
        j = ReadLELong(src + (i * 4)) & 0xFF ;
        palette[map.usedcol - 1 - i] |= (conv.gammac != 0)
                                      ? (USHORT)bmpgtab[j] << 8
                                      : (USHORT)(j & 0xF0) << 4 ;
    }

    /* Dfinit la couleur de tour pour TO-SNAP */
    src = file_buffer + bmp.bfOffBits ;
    map.outcol = GetMapBackGround(src) ;

    /* Remplit la table histogrammique */
    CreateHistogram(src) ;

    UnifyBitPlane(palette,src) ;

    /* Ajustements pour le bitmap 4 */
    if (map.video == 2)
    {
        /* Swappe les couleurs 1 et 2 de la palette */
        palette[4] = palette[1] ;
        palette[1] = palette[2] ;
        palette[2] = palette[4] ;
        /* Actualisation du tour d'cran */
        switch(map.outcol)
        {
             case 1 : map.outcol = 2 ; break ;
             case 2 : map.outcol = 1 ; break ;
        }
    }

    /* -------------------------------------------------------
     * Conversion des bitplans
     */

    /* Alloue la mmoire pour la conversion du fichier */
    planesize = map.width * map.height * 8 ;
    AllocTemp(planesize * 2) ;
    dst = temp_buffer ;
    if ((file_buffer == NULL) || (temp_buffer == NULL))
        return PError("",ERR_NOT_ENOUGH_MEMORY) ;

    for(width=0;width<map.width;width++)
    {
        offset = bmp.biSizeImage ;
        while (offset != 0L)
        {
            offset -= (bmp.biSizeImage / bmp.biHeight) ;
            switch(map.video)
            {
                /* Conversion pour 40 colonnes */
                case 0 :
                    ptr = src + (width << 2) + offset ;
                    get_bytes_40(palette, ptr, &c0, &c1) ;
                    *(dst+planesize) = c1 ;   /* Ecrit fond  */
                    *(dst++) = c0 ;           /* Ecrit forme */
                    break ;

                /* Conversion pour bitmap 4 */
                case 2 :
                    ptr = src + ((size_t)width << 2) + offset ;
                    /* Actualise la RAMB */
                    *(dst+planesize) = ~(((*ptr << 2) & 0x80)
                                      | ((*ptr << 5) & 0x40)
                                      | (*(ptr+1) & 0x20)
                                      | ((*(ptr+1) << 3) & 0x10)
                                      | ((*(ptr+2) >> 2) & 0x08)
                                      | ((*(ptr+2) << 1) & 0x04)
                                      | ((*(ptr+3) >> 4) & 0x02)
                                      | ((*(ptr+3) >> 1) & 0x01)) ;
                    /* Actualise la RAMA */
                    *(dst++) = ~(((*ptr << 3) & 0x80)
                                      | ((*ptr << 6) & 0x40)
                                      | ((*(ptr+1) << 1) & 0x20)
                                      | ((*(ptr+1) << 4) & 0x10)
                                      | ((*(ptr+2) >> 1) & 0x08)
                                      | ((*(ptr+2) << 2) & 0x04)
                                      | ((*(ptr+3) >> 3) & 0x02)
                                      | (*(ptr+3) & 0x01)) ;
                    break ;

                /* Conversion pour bitmap 16 et 80 colonnes */
                default :
                    *(dst++) = ~*(src + width + offset) ;
                    break ;
            }
        }
        /* Comble en hauteur avec la couleur du tour */
        for (i=bmp.biHeight;i<(ULONG)(map.height*8);i++)
        {
            switch(map.video)
            {
                case 0 : 
                    *(dst + planesize) = (map.outcol ^ 8) << 3 ;
                    *(dst++) = 0x00 ;
                    break ;
                case 1 :
                    *(dst++) = map.outcol * 0xFF ;
                    break ;
                case 2 :
                    *(dst + planesize) = (UBYTE)(((map.outcol & 1) != 0) ? 0xFF : 0x00) ;
                    *(dst++) = (UBYTE)(((map.outcol & 1) != 0) ? 0xFF : 0x00) ;
                    break ;
                case 3 :
                    *(dst++) = (map.outcol << 4) | map.outcol ;
                    break ;
            }
        }
    }

    /* Optimisation de la compression pour le 40 colonnes */
    if (map.video == 0)
    {
        OptimizeCmp40(temp_buffer,planesize) ;
    }

    /* -------------------------------------------------------
     * Cration du fichier MAP
     */

    /* Alloue la mmoire pour la cration du fichier MAP */
    FreeTemp((size_t)(dst - temp_buffer)) ;
    AllocTemp(65536 + 3 + 40) ;    
    if ((file_buffer == NULL) || (temp_buffer == NULL))
        return PError("",ERR_NOT_ENOUGH_MEMORY) ;
    src = file_buffer ;
    dst = temp_buffer ;

    /* Ecrit le header MAP */
    *(dst++) = map.scrmod ;            /* ScreenMode  */
    *(dst++) = (UBYTE)(map.width-1) ;  /* Largeur MAP */
    *(dst++) = (UBYTE)(map.height-1) ; /* Hauteur MAP */

    /* Compresse le corps du graphique
     * La compression RLE est optimise */
    for(i=0;i<2;i++)
    {  
        if ((map.video & i) == 0)
        {
            dst += CrunchMapPlane(src,dst,planesize) ;
            src += planesize ;
        }
        *(dst++) = 0x00 ;
        *(dst++) = 0x00 ;
    }


    /* Ajout de l'extension ventuelle */
    switch(maptype)
    {
       case CONV_GFX_SNP :
           /* Assure la parit d'adresse pour l'extension */
           if (((dst-temp_buffer) & 1) != 0) *(dst++) = 0x00 ;           
           WriteBEWord(dst,map.scrmod) ;   /* Ecrit le ScreenMode */
           WriteBEWord(dst+2,map.outcol) ; /* Ecrit la couleur du tour */
           WriteBEWord(dst+4,map.video) ;  /* Ecrit le VideoMode */
           for(i=0;i<16;i++)               /* Ecrit la palette TO-SNAP */
               WriteBEWord(dst+6+(i*2),palette[i]) ;
           WriteBEWord(dst+(16*2+6),0xa55a) ; /* Ecrit le marquage TO-SNAP */
           dst += 16*2+8 ;
           break ;

       case CONV_GFX_PPM :
           /* Assure la parit d'adresse pour l'extension */
           if (((dst-temp_buffer) & 1) != 0) *(dst++) = 0x00 ;
           for(i=0;i<16;i++)                   /* Ecrit la palette PPM */
               WriteBEWord(dst+(i*2),palette[15-i]) ;
           WriteBEWord(dst+(16*2),map.video) ; /* Ecrit le VideoMode */
           WriteBEWord(dst+(16*2+2),0x484C) ;  /* Ecrit le marquage PPM ("HL") */
           dst += 16*2+4 ;
           break ;
    }

    /* Officialisation des nouvelles adresse et taille de fichier */
    FreeTemp((size_t)(dst - temp_buffer)) ;

    /* Cre le fichier binaire */
    return RawToBin() ;
}


/*
 * Conversion d'un BMP en TO-SNAP
 */
int BmpToSnp(void)
{
    return BmpToMap(CONV_GFX_SNP) ;
}


/*
 * Conversion d'un BMP en PPM
 */
int BmpToPpm(void)
{
    return BmpToMap(CONV_GFX_PPM) ;
}




/*--------------------------------------------------------------------------
 * Traitement d'un fichier MAP (TO->PC)
 */

/*
 * Conversion d'une composante MAP en BMP
 */
static ULONG MapToBmpColor(ULONG color)
{
    color &= 0x0F ;
    return (conv.gammac != 0) ? mapgtab[color] & 0xFF : (color << 4) | color ;
}


/* ===========================================================================
 * Conversion MAP/PPM vers BMP
 * =========================================================================== */

int MapToBmp(void)
{
    ULONG i ;
    ULONG width, height, plane ;
    UBYTE value, digit1, digit2 ;
    ULONG planesize ;
    UBYTE *src ;
    UBYTE *dst ;
    UBYTE *ptr ;

    /* Palette BMP */
    ULONG bmpcmap[16] ;
    
    /* La palette est initialise avec les couleurs TO7-70 */
    USHORT palette[16] = {
        0x0000, 0x000f, 0x00f0, 0x00ff,
        0x0f00, 0x0f0f, 0x0ff0, 0x0fff,
        0x0777, 0x033a, 0x03a3, 0x03aa,
        0x0a33, 0x0a3a, 0x0ee7, 0x007b
    };

    memset(bmpcmap,0x00,sizeof(bmpcmap)) ;

    /*
     * Traitement des infos du fichier MAP
     */

    if (file_buffer == NULL)
        return PError("",ERR_NOT_ENOUGH_MEMORY) ;
    src = file_buffer ;

    /* Vrification de l'intgrit du fichier binaire
     * - Doivent tre  0x00 les octets : 0,3,4 et clture : 2,3,4,5
     * - 1er octet de clture  0xFF
     * - Taille du binaire correcte (binaire monobloc) */
    i = ReadBEWord(src+1) ;
    if (((i+10) != file_size)
     || ((*src | *(src+3) | *(src+4) | *(src+5+i+1) | *(src+5+i+2) | *(src+5+i+3) | *(src+5+i+4)) != 0x00)
     || (*(src+5+i) != 0xFF))
        return PError("",ERR_BAD_FILE_FORMAT);

    /* Rcupration des infos MAP */
    map.scrmod = (USHORT)*(src + 5)   ;   /* ScreenMode MAP */
    map.width  = (ULONG)(*(src + 6) + 1) ;  /* Largeur MAP    */
    map.height = (ULONG)(*(src + 7) + 1) ;  /* Hauteur MAP    */
    planesize  = map.width * map.height * 8 ; /* Taille d'un plan MAP */

    /* Initialisation des paramtres BMP selon le ScreenMode
     * Cette initialisation pralable est utile aussi pour
     * dfinir une palette par dfaut selon le mode d'cran,
     * mme quand elle n'est pas dtermine par une extension
     * TO-SNAP ou PPM */
    map.video          = 0 ;                      /* Mode vido 40 colonnes */
    bmp.biWidth        = (ULONG)(map.width * 8) ; /* Largeur bitmap         */
    bmp.biBitCount     = 4 ;                      /* Chunky 4 bits          */
    bmp.biClrUsed      = 16 ;                     /* Palette de 16 couleurs */
    bmp.biClrImportant = 16 ;                     /* 16 couleurs utilises  */

    switch(map.scrmod)
    {
        /* Ajustement pour 80 colonnes */
        case 0x80 :
            map.video = 1 ;          /* Mode vido 80 colonnes */
            palette[0] = 0x0fff ;    /* Couleur 0 blanche      */
            palette[1] = 0x0000 ;    /* Couleur 1 noire        */
            bmp.biBitCount = 1 ;     /* Chunky 1 bit           */
            bmp.biClrUsed = 2 ;      /* Palette de 2 couleurs  */
            bmp.biClrImportant = 2 ; /* 2 couleurs utilises   */
            break ;

        /* Ajustement pour bitmap 16 */
        case 0x40 :
            map.video = 3 ;
            bmp.biWidth = (ULONG)(map.width * 2) ;
            for(i=0;i<16;i++)
                palette[i] = i * 0x111 ;
            break ;

        /* Ajustement pour bitmap 4 et 40 colonnes (ne fait rien) */
        case 0x00 :
            break ;

        /* Erreur si autre format */
        default :
            return PError("",ERR_RESOLUTION) ;
    }
    bmp.biHeight = (ULONG)(map.height * 8) ;

    /* Recherche des extensions TO-SNAP ou PPM
     * Le marquage TO-SNAP et PPM se trouve au dernier mot
     * du fichier */
    src = file_buffer + file_size - 5 - 2 ;
    switch (ReadBEWord(src))
    {
        /* Marquage TO-SNAP */
        case 0xa55a :
            for(i=0;i<16;i++)   /* Rcupre palette TO-SNAP */
            {
                src -= 2 ;
                palette[15-i] = ReadBEWord(src) ;
                if ((palette[15-i] & 0x8000) != 0)
                    palette[15-i] = ((~palette[15-i]) & 0x0fff) ;
            }
            map.video  = ReadBEWord(src-2) ; /* Rcupre VideoMode */
            map.outcol = ReadBEWord(src-4) ; /* Rcupre couleur de tour */
            map.scrmod = ReadBEWord(src-6) ; /* Rcupre le ScreenMode */
            break ;

        /* Marquage PPM */
        case 0x484c :
            src -= 2 ;
            map.video = ReadBEWord(src) ; /* Rcupre VideoMode */
            for(i=0;i<16;i++)              /* Rcupre palette PPM */
            {
                src -= 2 ;
                palette[i] = ReadBEWord(src) ;
                if ((palette[i] & 0x8000) != 0)
                    palette[i] = ((~palette[i]) & 0x0fff) ;
            }
            break ;
    }

    /* Actualisation des paramtres du header BMP
     * (swap des couleurs 1 et 2 pour bitmap 4) */
    if (map.video == 2) 
    {
        /* Swappe les couleurs 1 et 2 */
        palette[15] = palette[1] ;
        palette[1] = palette[2] ;
        palette[2] = palette[15] ;
        /* Actualise le nombre de couleurs utilises */
        bmp.biClrImportant = 4 ;
    }

    /* Conversion de la palette pour le fichier BMP
     * La complmentation des bitplans est compense par le swap
     * de la palette
     * La gamma correction de la palette est effectue si
     * demande */
    for(i=0;i<bmp.biClrImportant;i++)
        bmpcmap[bmp.biClrImportant - 1 -i] = MapToBmpColor((ULONG)palette[i] >> 8)
                                           | (MapToBmpColor((ULONG)palette[i] >> 4) << 8)
                                           | (MapToBmpColor((ULONG)palette[i]) << 16) ;

    /* Alloue la mmoire pour la conversion du fichier */
    file_size = planesize * (2 - (map.video & 1)) ;
    AllocTemp((size_t)file_size) ;
    if ((file_buffer == NULL) || (temp_buffer == NULL))
        return PError("",ERR_NOT_ENOUGH_MEMORY) ;
    dst = temp_buffer ;

    /* Dcompression du corps du graphique
     * L'intgrit du fichier MAP/PPM est vrifie */
    plane = 0 ;   /* Nombre de passages  0 */
    i = 0 ;       /* Compteur de taille de bitplan  0 */
    src = file_buffer + 5 + 3 ;
    while(plane < 2)
    {
        /* Squence d'galit */
        if ((value = *(src++)) != 0)
        {
            /* Erreur si risque d'erreur de segmentation */
            if ((size_t)((temp_buffer - dst) + value) > file_size)
                return PError("",ERR_COMPRESSION) ;

            /* Remplit la destination */
            memset(dst,(int)*src,(size_t)value) ;
            dst += value ;
            src++ ;
            i += value ;
        }
        else
        {
            /* Squence d'ingalit */
            if ((value = *(src++)) != 0)
            {
                /* Erreur si risque d'erreur de segmentation */
                if ((size_t)((temp_buffer - dst) + value) > file_size)
                    return PError("",ERR_COMPRESSION) ;
                /* Remplit la destination */
                memcpy(dst,src,(size_t)value) ;
                dst += value ;
                src += value ;
                i += value ;
            }
            else
            {
                /* Fin d'un bitplan
                 * Erreur si :
                 * - bitmap 4 ou 40 colonnes et taille incorrecte
                 * - bitmap 16 ou 80 colonnes et taille incorrecte
                 * - bitmap 16 ou 80 colonnes et 2me plan non vide */
                if (((map.scrmod == 0) && (i != planesize))
                 || ((map.scrmod != 0) && (((plane == 0) && (i != planesize))
                                   || ((plane != 0) && (i != 0)))))
                    return PError("",ERR_COMPRESSION) ;
                i = 0 ;    /* Compteur de taille de bitplan  0 */
                plane++ ;  /* Augmente numro de bitplan */
            }
        }
    }

    /* Actualisation de l'en-tte BMP */
    bmp.bfOffBits   = BITMAPHEADERSIZE + (bmp.biClrUsed * 4) ;
    bmp.biSizeImage = ((((bmp.biWidth * bmp.biBitCount) + 31) & 0xFFFFFFE0) >> 3) * bmp.biHeight ;
    bmp.bfSize      = bmp.bfOffBits + bmp.biSizeImage ;

    /* Alloue la mmoire pour la cration du fichier BMP */
    FreeTemp((size_t)(dst - temp_buffer)) ;
    AllocTemp((size_t)bmp.bfSize) ;
    if ((file_buffer == NULL) || (temp_buffer == NULL))
        return PError("",ERR_NOT_ENOUGH_MEMORY) ;
    src = file_buffer ;
    dst = temp_buffer ;

    /* Enregistrement de l'en-tte BMP */
    WriteBEWord(dst,   16973) ;              /* bfType            */
    WriteLELong(dst+2 ,bmp.bfSize) ;         /* bfSize            */
    WriteLEWord(dst+6 ,0) ;                  /* bfReserved1       */
    WriteLEWord(dst+8 ,0) ;                  /* bfReserved2       */
    WriteLELong(dst+10,bmp.bfOffBits) ;      /* bfOffBits         */
    WriteLELong(dst+14,40) ;                 /* biSize            */
    WriteLELong(dst+18,bmp.biWidth) ;        /* biWidth           */
    WriteLELong(dst+22,bmp.biHeight) ;       /* biHeight          */
    WriteLEWord(dst+26,1) ;                  /* biPlanes          */
    WriteLEWord(dst+28,bmp.biBitCount) ;     /* biBitCount        */
    WriteLELong(dst+30,0) ;                  /* biCompression     */
    WriteLELong(dst+34,bmp.biSizeImage) ;    /* biSizeImage       */
    WriteLELong(dst+38,4875) ;               /* biXPixelsPerMeter */
    WriteLELong(dst+42,4875) ;               /* biYPixelsPerMeter */
    WriteLELong(dst+46,bmp.biClrUsed) ;      /* biClrUsed         */
    WriteLELong(dst+50,bmp.biClrImportant) ; /* biClrImportant    */

    /* Enregistrement de la palette */
    for(i=0;i<bmp.biClrUsed;i++)
        WriteLELong(dst + BITMAPHEADERSIZE + (i * 4),(ULONG)bmpcmap[i]) ;

    /* Enregistrement de la bitmap */
    dst += bmp.bfOffBits ;
    height = map.height * 8 ;
    while(height != 0)
    {
        height-- ;
        for(width=0;width<planesize;width+=map.height*8)
        {
            ptr = src + width + height ;
            switch(map.video)
            {
                /* Conversion pour 40 colonnes */
                case 0 :
                    /* Lit l'octet RAMB */
                    value = ~*(ptr + planesize) ;
                    /* Dtermine couleurs de fond et de forme */
                    digit1 = ((value & 0x78) >> 3) ^ 0x08 ;
                    digit2 = (((value & 0x80) >> 4) | (value & 0x07)) ^ 0x08 ;
                    /* Lit l'octet RAMA */
                    value = ~*ptr ;
                    /* Ecrit la correspondance couleur */
                    *(dst++) = ((((value & 0x80) != 0) ? digit2 : digit1) << 4)
                              | (((value & 0x40) != 0) ? digit2 : digit1) ;
                    *(dst++) = ((((value & 0x20) != 0) ? digit2 : digit1) << 4)
                              | (((value & 0x10) != 0) ? digit2 : digit1) ;
                    *(dst++) = ((((value & 0x08) != 0) ? digit2 : digit1) << 4)
                              | (((value & 0x04) != 0) ? digit2 : digit1) ;
                    *(dst++) = ((((value & 0x02) != 0) ? digit2 : digit1) << 4)
                              | (((value & 0x01) != 0) ? digit2 : digit1) ;
                    break ;

                /* Conversion pour bitmap 4 */
                case 2 :
                    /* Lit octets RAMA et RAMB */
                    digit1 = ~*ptr ;
                    digit2 = ~*(ptr + planesize) ;
                    /* Ecrit la correspondance couleur */
                    *(dst++) = ((digit1 & 0x80) >> 3) | ((digit2 & 0x80) >> 2)
                             | ((digit1 & 0x40) >> 6) | ((digit2 & 0x40) >> 5) ;
                    *(dst++) = ((digit1 & 0x20) >> 1) | (digit2 & 0x20)
                             | ((digit1 & 0x10) >> 4) | ((digit2 & 0x10) >> 3) ;
                    *(dst++) = ((digit1 & 0x08) << 1) | ((digit2 & 0x08) << 2)
                             | ((digit1 & 0x04) >> 2) | ((digit2 & 0x04) >> 1) ;
                    *(dst++) = ((digit1 & 0x02) << 3) | ((digit2 & 0x02) << 4)
                             | (digit1 & 0x01) | ((digit2 & 0x01) << 1) ;
                    break;

                /* Conversion bitmap 16 et 80 */
                default :
                    *(dst++) = ~*ptr ;
                    break ;
            }
        }
        /* Comble la ligne graphique BMP */
        width = (bmp.biWidth * bmp.biBitCount) ;
        dst += (((width + 31) & 0xFFFFFFE0) >> 3) - (width >> 3) ;
    }

    /* Officialisation des nouvelles adresse et taille de fichier */
    FreeTemp((size_t)(dst - temp_buffer)) ;

    return TRUE ;
}
