/*       ___   ___   ___   ___
 *      /_  \ /   \ /  _\ /  _\
 *     // \_/// \_/_\ / \\  / \\
 *    //  _ //  _ / \\\_// /  //
 *    \\_/ \\\_/ \\ //  /  \_//
 *     \___/ \___/ \\__/ \___/
 *     Transferts srie pour Thomson
 *
 *  Dfinitions
 *
 *  Copyright (C) 2015 Puls
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <dirent.h>
#include <stdint.h>

#define CC90_VERSION0 "1"
#define CC90_VERSION1 "0"
#define CC90_VERSION2 "4"
#define CC90_DATE     "mars 2015"


#ifndef DEFS_H
#define DEFS_H

#ifndef TRUE
#   define TRUE 1
#endif

#ifndef FALSE
#   define FALSE 0
#endif

#ifndef NULL
#   define NULL 0
#endif


#ifdef _GNU_SOURCE
#   define SYSTEMSLASH "/"
#else
#   define SYSTEMSLASH "\\"
#endif

#define STRING_SIZE_MAX  300

/* Codes de transfert srie */
#define CODE_BREAK         0
#define CODE_ASK           1
#define CODE_LOADSECTORS   2
#define CODE_SAVESECTORS   3
#define CODE_SENDMEMORY    4
#define CODE_GETMEMORY     5
#define CODE_FORMATDISK    6

/* Codes contrleurs */
#define TO_COMPUTER    0x08
#define MO5_COMPUTER   0x00
#define MO6_COMPUTER   0x01
#define TO7_COMPUTER   0x00+TO_COMPUTER
#define TO770_COMPUTER 0x01+TO_COMPUTER
#define TO9_COMPUTER   0x02+TO_COMPUTER
#define TO8_COMPUTER   0x03+TO_COMPUTER
#define TO9P_COMPUTER  0x06+TO_COMPUTER

#define CURRENT_BANK 0

#define CHAR127  '|'

#define MAXSECSIZE        256
#define SECTORS_PER_TRACK 16
#define TRACKSIZE         SECTORS_PER_TRACK*MAXSECSIZE
#define DIRTRACK   20
#define FATSECTOR  2
#define MAXFATSIZE 160
#define ENTRYSIZE  32
#define FIRSTB1 (UBYTE)1
#define LASTB1  (UBYTE)8
#define FIRSTB2 (UBYTE)9
#define LASTB2  (UBYTE)16

#define MIN_WAIT_TIME   1500
#define MID_WAIT_TIME   -1
#define MAX_WAIT_TIME   -2


#define  DRIVE0   0
#define  DRIVE1   1
#define  DRIVE2   2
#define  DRIVE3   3

typedef uint8_t UBYTE ;
typedef uint16_t USHORT ;
typedef uint32_t ULONG ;

enum {
    INVALID_TYPE,
    TO_TYPE,
    SAP_TYPE,
    DCMOTO_TYPE,
    FILE_TYPE
} ;

enum {
    OVERWRITE_FLAG,
    DELETE_FLAG
} ;

struct DISK_T {
    char   arg[STRING_SIZE_MAX] ;
    UBYTE  drive ;
    char   dkname[STRING_SIZE_MAX] ;
    char   pattern[STRING_SIZE_MAX] ;
    char   filename[STRING_SIZE_MAX] ;
    DIR    /*@dependent@*//*@null@*/*drbase ;       /* Handler DIR principal */
    int    type ;
    int    dir ;
    int    cont ;
    UBYTE  nbtracks ;
    size_t fatsize ;
    size_t secsize ;
    int    external ;
    /* Pour les opration systme */
    UBYTE  /*@dependent@*/*fat ;
    UBYTE  *tk20 ;
    UBYTE  /*@dependent@*/*entry ;
    UBYTE  /*@dependent@*//*@null@*/*free ;
    
} ;

struct CC90_INFO_T {
    UBYTE  code ;
    UBYTE  data[5] ;
    UBYTE  crchigh ;
    UBYTE  crclow ;
} ;

struct CC90_T {
    char   portname[STRING_SIZE_MAX] ;
    int    timeout_max ;
    int    timeout_mid ;
    int    timeout ;
    struct CC90_INFO_T info ;
    UBYTE  synchro[2] ;
    USHORT ptrack ;
    UBYTE  config ;
    UBYTE  lect ;
    char   iddisk ;
    UBYTE  diskop ;
    int    lasterror ;
    int    yes ;
    int    exit ;
} ;

/* Paramtrage commande */
struct OPT_T {
    int   quick ;
    int   interlace ;
    UBYTE bank ;
    int   autochg ;
} ;


enum {
    CONV_GFX_SNP,
    CONV_GFX_PPM
} ;

enum {
    CONV_SND_SON,
    CONV_SND_SND
} ;

enum {
    CONV_TXT_ASM,
    CONV_TXT_ASC
} ;


/* Paramtrage conversions */
struct CONV_T {
    int     convert ;
    ULONG   sample_rate ;
    int     tabul ;
    int     keymap_number ;
    int     maptype ;    /* CONV_GFX_SNP|CONV_GFX_PPM|CONV_GFX_MAP */
    int     sndtype ;    /* CONV_SND_SND|CONV_SND_SON */
    int     txttype ;    /* CONV_TXT_ASM|CONV_TST_ASC */
    int     videomode ;  /* 0=40cols 1=80cols 2=bitmap4  3=bitmap16 */
    int     force40cols ; /* TRUE = Force graphiques en 40 colonnes */
    int     gammac ;     /* Switch pour gamma correction */
} ;


struct CONTROLLER {
    UBYTE   to ;
    UBYTE   sap ;
    long    dcmoto ;
    char    dcmotosuffix[3] ;
    UBYTE   maxdrive ;
    UBYTE   nbtracks ;
    size_t  fatsize ;
    size_t  secsize ;
} ;


/* Liste des erreurs */
enum {
    /* Les erreurs suivantes interrompent
     * immdiatement une opration avec
     * wildcards */
    ERR_PORT_NOT_READY = -100,
    ERR_BREAK,
    ERR_SERIAL_ERROR,
    ERR_NOT_ENOUGH_MEMORY,
    ERR_CRC_ERROR,
    ERR_SAME_DISKS,
    ERR_ILLEGAL_WILDCARDS,
    ERR_FATFULL,
    ERR_FILE_NOT_FOUND,
    ERR_DIR_NOT_FOUND,
    ERR_DIRECTORY_FULL,
    ERR_CANT_OPEN_FILE,
    ERR_BAD_FAT,
    ERR_BAD_ENTRY,
    ERR_NO_DISK,
    ERR_IO_ERROR,
    ERR_INCOMPATIBLE_DISK,
    ERR_ILLEGAL_OPERATION,
    ERR_COPY_DIR_INTO_FILE,
    ERR_FILE_IS_A_DIR,
    ERR_KEY_NOT_FOUND,
    ERR_SPACE_AFTER_KEY,
    ERR_EQUAL_AFTER_KEY,
    ERR_SECTION_NOT_FOUND,
    ERR_UNKNOWN_VALUE,
    ERR_VALUE_NOT_DECIMAL,
    /* Les erreurs suivantes n'interrompent
     * pas une opration avec wildcards */
    ____DONT_STOP,
    ERR_PERMISSION_DENIED,
    ERR_BAD_FILE_NAME,
    ERR_EMPTY_FILE_NAME,
    ERR_BAD_FILE_FORMAT,
    ERR_BIN_FILE_TOO_LONG,
    ERR_RESOLUTION,
    ERR_COMPRESSION
} ;

#endif
