/*       ___   ___   ___   ___
 *      /_  \ /   \ /  _\ /  _\
 *     // \_/// \_/_\ / \\  / \\
 *    //  _ //  _ / \\\_// /  //
 *    \\_/ \\\_/ \\ //  /  \_//
 *     \___/ \___/ \\__/ \___/
 *     Transferts srie pour Thomson
 *
 *  Oprations fichiers
 *
 *  Copyright (C) 2015 Puls
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/types.h>
#include <dirent.h>
#include <errno.h>

#include "defs.h"

extern UBYTE *file_buffer ;
extern ULONG file_size ;

extern void AllocFile(size_t size) ;
extern void FreeFile(void) ;

extern FILE /*@null@*//*@dependent@*/*fp_file ;

extern int PError(char *, int) ;


/*
 * Fermeture d'un fichier
 */
int /*@unused@*/
CloseFpFile(void) /*@globals errno@*/
{
    if (fp_file != NULL) 
        if (fclose(fp_file) != 0)
            return FALSE ;
    fp_file = NULL ;
    return TRUE ;
}


/*
 * Ouverture d'un fichier
 */
int /*@unused@*/ 
OpenFpFile(int create, char *name, char *mode) /*@globals errno@*/
{
    if ((fp_file = fopen(name,mode)) == NULL)
    {
        switch(errno)
        {
            case EISDIR :
                 return PError(name,ERR_FILE_IS_A_DIR) ;
            case EACCES :
            case EROFS :
                 return PError(name,ERR_PERMISSION_DENIED) ;
            case ENAMETOOLONG :
                 return PError(name,ERR_BAD_FILE_NAME) ;
            case ENOENT :
                 if (create == TRUE)
                     return ERR_DIR_NOT_FOUND ;
                 return PError(name,ERR_DIR_NOT_FOUND) ;
            case ENOTDIR :
            case ENXIO :
                 if (create == TRUE)
                     return ERR_FILE_NOT_FOUND ;
                 return PError(name,ERR_FILE_NOT_FOUND) ;
            default :
                 return PError(name,ERR_CANT_OPEN_FILE) ;
        }
    }
    return TRUE ;
}


/*
 * Rcupre la taille d'un fichier
 */
int /*@unused@*/
FpFileSize(char *name) /*@globals errno@*/
{
    long int pos = 0 ;
    if (OpenFpFile(FALSE,name,"rb") == FALSE)
        return FALSE ;
    if (fp_file == NULL)
        return PError(name,ERR_IO_ERROR) ;

    if ((fseek(fp_file, 0, SEEK_END) != 0)
     || ((pos = ftell(fp_file)) < 0))
    {
        (void)CloseFpFile() ;
        return PError(name,ERR_IO_ERROR) ;
    }
    if (CloseFpFile() == FALSE)
        return PError(name,ERR_IO_ERROR) ;
    file_size = (size_t)pos ;
    return TRUE ;
}


/*
 * Charge un fichier en mmoire
 */
int
LoadFpFile(char *name) /*@globals errno@*/
{
    /* Rcupre la taille du fichier */
    if (FpFileSize(name) == FALSE) return FALSE ;

    /* Alloue la mmoire pour le chargement
     * La dclaration est lgrement suprieure
     *  la taille relle pour prvenir les
     * dbordements de remplissage ou de
     * conversion */
    AllocFile(file_size) ;

    /* Charge le fichier */
    if (OpenFpFile(FALSE,name,"rb") == FALSE)
    {
        FreeFile() ;
        return FALSE ;
    }
    if (fp_file == NULL)
        return PError(name,ERR_IO_ERROR) ;
    if (fread(file_buffer, sizeof(UBYTE), file_size, fp_file) != file_size)
    {
        FreeFile() ;
        (void)CloseFpFile() ;
        return PError(name,ERR_IO_ERROR) ;
    }
    (void)CloseFpFile() ;

    return TRUE ;
}

