/*       ___   ___   ___   ___
 *      /_  \ /   \ /  _\ /  _\
 *     // \_/// \_/_\ / \\  / \\
 *    //  _ //  _ / \\\_// /  //
 *    \\_/ \\\_/ \\ //  /  \_//
 *     \___/ \___/ \\__/ \___/
 *     Transferts srie pour Thomson
 *
 *  Oprations mmoire
 *
 *  Copyright (C) 2015 Puls
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <ctype.h>

#include "defs.h"

UBYTE /*@null@*//*@only@*//*@out@*/*track_buffer = NULL ;
UBYTE /*@null@*//*@only@*//*@out@*/*file_buffer = NULL ;
UBYTE /*@null@*//*@only@*//*@out@*/*dos_buffer = NULL ;
UBYTE /*@null@*//*@only@*//*@out@*/*temp_buffer = NULL ;
ULONG file_size = 0 ;
ULONG temp_size = 0 ;

extern int PError(char *, int) ;
extern int CloseAll(void) ;

/*
 * Libre la mmoire du fichier
 */
void /*@unused@*/ FreeFile(void)
{
    if (file_buffer != NULL) {
        free(file_buffer) ;
        file_buffer = NULL ;
    }
}


/*
 * Libre toutes les mmoires
 */
void /*@unused@*/ FreeAll(void)
{
    FreeFile() ;

    if (track_buffer != NULL) {
        free(track_buffer) ;
        track_buffer = NULL ;
    }
    if (dos_buffer != NULL) {
        free(dos_buffer) ;
        dos_buffer = NULL ;
    }
    if (temp_buffer != NULL) {
        free(temp_buffer) ;
        temp_buffer = NULL ;
    }
}


/*
 * Sort si erreur d'allocation
 */
static void AllocError(void)
{
    FreeAll() ;
    (void)PError("", ERR_NOT_ENOUGH_MEMORY) ;
    exit(CloseAll()) ;
}


/*
 * Alloue les buffers de pistes
 */
int AllocTracks(void)
{
    if (track_buffer != NULL) {
        free(track_buffer) ;
        track_buffer = NULL ;
    }
    track_buffer = malloc(TRACKSIZE*2) ;
    if (track_buffer == NULL) {
        AllocError() ;
        return FALSE ;
    }
    memset(track_buffer,0xE5,TRACKSIZE*2) ;

    if (dos_buffer != NULL) {
        free(dos_buffer) ;
        dos_buffer = NULL ;
    }
    dos_buffer = malloc(TRACKSIZE*4) ;
    if (dos_buffer == NULL) {
        AllocError() ;
        return FALSE ;
    }
    memset(dos_buffer,0xFF,TRACKSIZE*4) ;

    return TRUE ;
}


/*
 * Alloue la mmoire pour le fichier
 */
void AllocFile(size_t size)
{
    if (file_buffer != NULL) {
        free(file_buffer) ;
        file_buffer = NULL ;
    }
    file_buffer = malloc(size) ;
    if (file_buffer == NULL) {
        AllocError() ;
        return ;
    }
    memset(file_buffer,0x00,size) ;
}


/*
 * Alloue la mmoire pour le buffer temporaire
 */
void AllocTemp(size_t size)
{
    if (temp_buffer != NULL) {
        free(temp_buffer) ;
        temp_buffer = NULL ;
    }
    temp_buffer = malloc(size) ;
    if (temp_buffer == NULL) {
        AllocError() ;
        return ;
    }
    memset(temp_buffer,0x00,size) ;
    temp_size = size ;
}


void IncreaseTempBuffer(size_t size)
{
    UBYTE /*@only@*//*@null@*/ *t ;
    
    t = malloc(temp_size + size) ;
    if ((t != NULL) && (temp_buffer != NULL))
    {
        memmove(t,(UBYTE /*@temp@*/ *)temp_buffer,temp_size) ;
        temp_size += size ;
        if (temp_buffer != NULL) {
            free(temp_buffer) ;
            temp_buffer = NULL ;
        }
        temp_buffer = t ;
        t = NULL ;
    } else
        AllocError() ;
}

/*
 * Met les  fichier  jour
 */
void FreeTemp(size_t size)
{
    if (file_buffer != NULL) {
        free(file_buffer) ;
        file_buffer = NULL ;
    }
    file_size = size ;
    file_buffer = temp_buffer ;
    temp_buffer = NULL ;
    temp_size = 0 ;
}

